import datetime
from dateutil.relativedelta import relativedelta

from ldap3.protocol.formatters.formatters import format_sid
from impacket.ldap.ldaptypes import SR_SECURITY_DESCRIPTOR
from impacket.uuid import bin_to_string

from powerview.modules.msa import MSA
from powerview.utils.constants import (
	UAC_DICT,
	LDAP_ERROR_STATUS,
	SUPPORTED_sAMAccountType,
	SUPPORTED_ENCRYPTION_TYPES,
	switcher_trustDirection,
	switcher_trustType,
	switcher_trustAttributes,
	PWD_FLAGS,
	FOREST_TRUST_INFO,
	switcher_dsa_delegated_msa_state
)
from powerview.modules.ldapattack import (
	RBCD
)
from powerview.utils.helpers import filetime_to_str

class UAC:
	@staticmethod
	def parse_value(uac_value):
		uac_value = int(uac_value)
		flags = []

		for key, value in UAC_DICT.items():
			if uac_value & key:
				flags.append(value)

		return flags

	@staticmethod
	def parse_value_tolist(uac_value):
		uac_value = int(uac_value)
		flags = []

		for key, value in UAC_DICT.items():
			if uac_value & key:
				flags.append([value, key])

		return flags

	@staticmethod
	def parse_uac_namestrings_to_value(uac_names, type=list):
		"""Convert UAC flag names to numeric value
		Args:
			uac_names (str): UAC flag names
		Returns:
			int: Combined UAC numeric value
		"""
		if not uac_names:
			return 0
		if isinstance(uac_names, str):
			names = [name.strip() for name in uac_names.split(",")]
		elif isinstance(uac_names, list):
			names = [name.strip() for name in uac_names]
		else:
			raise TypeError("uac_names must be a string or a list")

		uac_value = 0
		reverse_uac = {value: key for key, value in UAC_DICT.items()}

		for name in names:
			if name in reverse_uac:
				uac_value |= reverse_uac[name]
			else:
				raise ValueError(f"Invalid UAC name: {name}")
		
		if type == list:
			return [uac_value]
		elif type == int:
			return int(uac_value)
		else:
			return uac_value

class ENCRYPTION_TYPE:
	@staticmethod
	def parse_value(enc_value):
		enc_value = int(enc_value)
		flags = []

		for key, value in SUPPORTED_ENCRYPTION_TYPES.items():
			if enc_value & key:
				flags.append(value)

		return flags

class sAMAccountType:
	@staticmethod
	def parse_value(enc_value):
		enc_value = int(enc_value)

		if enc_value in SUPPORTED_sAMAccountType:
			return SUPPORTED_sAMAccountType[enc_value]
		else:
			return enc_value

class LDAP:
	@staticmethod
	def resolve_pKIExpirationPeriod(data):
		try:
			return filetime_to_str(data)
		except Exception as e:
			return data

	@staticmethod
	def resolve_delegated_msa_state(data):
		if isinstance(data, list):
			return switcher_dsa_delegated_msa_state.get(int(data[0]))
		elif isinstance(data, bytes):
			return switcher_dsa_delegated_msa_state.get(int(data.decode()))
		else:
			return switcher_dsa_delegated_msa_state.get(int(data))

	@staticmethod
	def resolve_pKIOverlapPeriod(data):
		try:
			return filetime_to_str(data)
		except Exception as e:
			return data

	@staticmethod
	def resolve_msDSTrustForestTrustInfo(data):
		sids = []
		parser = FOREST_TRUST_INFO(data)
		for record in parser['Records']:
			try:
				sids.append(record['Data']['Sid'].formatCanonical())
			except KeyError:
				pass
		return sids

	@staticmethod
	def resolve_msDSAllowedToActOnBehalfOfOtherIdentity(data):
		sids = []
		sd = SR_SECURITY_DESCRIPTOR(data=data)
		if len(sd['Dacl'].aces) > 0:
			for ace in sd['Dacl'].aces:
				sids.append(ace['Ace']['Sid'].formatCanonical())
			return sids
		else:
			return data

	@staticmethod
	def resolve_err_status(error_status):
		return LDAP_ERROR_STATUS.get(error_status)

	@staticmethod
	def resolve_enc_type(enc_type):
		if isinstance(enc_type, list):
			return ENCRYPTION_TYPE.parse_value(enc_type[0])
		elif isinstance(enc_type, bytes):
			return ENCRYPTION_TYPE.parse_value(enc_type.decode())
		else:
			return ENCRYPTION_TYPE.parse_value(enc_type)

	@staticmethod
	def resolve_samaccounttype(enc_type):
		if isinstance(enc_type, list):
			return sAMAccountType.parse_value(enc_type[0])
		elif isinstance(enc_type, bytes):
			return sAMAccountType.parse_value(enc_type.decode())
		else:
			return sAMAccountType.parse_value(enc_type)

	@staticmethod
	def resolve_uac(uac_val):
		# resolve userAccountControl
		if isinstance(uac_val, list):
			val =  UAC.parse_value(uac_val[0])
		elif isinstance(uac_val, bytes):
			val = UAC.parse_value(uac_val.decode())
		else:
			val = UAC.parse_value(uac_val)

		return val

	@staticmethod
	def parse_uac_name_to_value(uac_names, delimiter=','):
		"""Convert UAC flag names to numeric value
		Args:
			uac_names (str): Comma-separated UAC flag names
		Returns:
			int: Combined UAC numeric value
		"""
		if not uac_names:
			return 0
			
		uac_value = 0
		names = [name.strip() for name in uac_names.split(delimiter)]
		
		# Create reverse mapping of name->value
		reverse_uac = {value: key for key, value in UAC_DICT.items()}
		
		for name in names:
			if name in reverse_uac:
				uac_value |= reverse_uac[name]
				
		return uac_value

	@staticmethod
	def ldap2datetime(ts):
		if isinstance(ts, datetime.datetime):
			return ts
		else:
			ts = int(ts)
			dt = datetime.datetime(1601, 1, 1) + datetime.timedelta(seconds=ts / 10000000)
		#return datetime.datetime(1601, 1, 1) + datetime.timedelta(seconds=ts/10000000)
		return f"{dt.strftime('%d/%m/%Y %H:%M:%S')} ({LDAP.human_readable_time_diff(dt)})"

	@staticmethod
	def human_readable_time_diff(past_date):
		now = datetime.datetime.now()
		diff = relativedelta(now, past_date)

		if diff.years > 0:
			return f"{diff.years} year{'s' if diff.years > 1 else ''}, {diff.months} month{'s' if diff.months > 1 else ''} ago"
		elif diff.months > 0:
			return f"{diff.months} month{'s' if diff.months > 1 else ''}, {diff.days} day{'s' if diff.days > 1 else ''} ago"
		elif diff.days > 0:
			return f"{diff.days} day{'s' if diff.days > 1 else ''} ago"
		else:
			return "today"

	@staticmethod
	def resolve_generalized_time(ldap_time):
		if isinstance(ldap_time, datetime.datetime):
			dt = ldap_time
		else:
			if isinstance(ldap_time, bytes):
				ldap_time = ldap_time.decode()
			dt = datetime.datetime.strptime(ldap_time, "%Y%m%d%H%M%S.%fZ")
		
		return f"{dt.strftime('%d/%m/%Y %H:%M:%S')} ({LDAP.human_readable_time_diff(dt)})"

	@staticmethod
	def bin_to_guid(guid):
		return "{%s}" % bin_to_string(guid).lower()

	@staticmethod
	def bin_to_sid(sid):
		return format_sid(sid)

	@staticmethod
	def formatGMSApass(managedPassword):
		return MSA.decrypt(managedPassword)

	@staticmethod
	def parseMSAMembership(secDesc):
		return MSA.read_acl(secDesc)

	@staticmethod
	def resolve_pwdProperties(flag):
		prop =  PWD_FLAGS.get(int(flag))
		return f"({flag.decode()}) {prop}" if prop else flag

class TRUST:
	@staticmethod
	def resolve_trustDirection(flag):
		flag = int(flag)
		types = []
		for bit, name in switcher_trustDirection.items():
			if flag & bit:
				types.append(name)
		return types

	@staticmethod
	def resolve_trustType(flag):
		flag = int(flag)
		types = []
		for bit, name in switcher_trustType.items():
			if flag & bit:
				types.append(name)
		return types

	@staticmethod
	def resolve_trustAttributes(flag):
		flag = int(flag)
		attributes = []
		for bit, name in switcher_trustAttributes.items():
			if flag & bit:
				attributes.append(name)
		return attributes

class EXCHANGE:
	@staticmethod
	def resolve_msExchVersion(version_num):
		"""
		Convert an msExchVersion number to human-readable Exchange version
		
		Args:
			version_num: The msExchVersion value as a string or integer
			
		Returns:
			String with decoded version information
		"""
		try:
			if isinstance(version_num, list):
				version_num = version_num[0]
			elif isinstance(version_num, bytes):
				version_num = version_num.decode()
				
			version_num = int(version_num)
			
			# Extract components
			major_version = (version_num >> 48) & 0xFFFF
			minor_version = (version_num >> 32) & 0xFFFF
			build = (version_num >> 16) & 0xFFFF
			revision = version_num & 0xFFFF
			
			# Determine Exchange version
			exchange_version = ""
			if major_version == 15:
				if minor_version >= 2:
					exchange_version = "Exchange 2019"
				elif minor_version == 1:
					exchange_version = "Exchange 2016"
				elif minor_version == 0:
					exchange_version = "Exchange 2013"
			elif major_version == 14:
				exchange_version = "Exchange 2010"
			elif major_version == 8:
				exchange_version = "Exchange 2007"
			elif major_version == 6:
				exchange_version = "Exchange 2003"
			else:
				exchange_version = f"Unknown Exchange (Version {major_version})"
			
			version_str = f"{major_version}.{minor_version}.{build}.{revision}"
			return f"{exchange_version} ({version_str})"
		except (ValueError, TypeError) as e:
			return f"Unknown format: {version_num}"



from ..templates import LDAP_ADD_FSTRING

from ldap3 import SEQUENCE_TYPES
from ldap3.protocol.rfc4511 import LDAPDN
from uuid import uuid4

ATTRIBUTE_TYPE_AND_VALUE_FSTRING = """<AttributeTypeAndValue>
    <AttributeType>{attribute}</AttributeType>
    <AttributeValue>
        <ad:value xsi:type="xsd:string">{value}</ad:value>
    </AttributeValue>
</AttributeTypeAndValue>"""

def get_rdn(dn: str):
    return dn.split(',')[0]

def get_parent_dn(dn: str):
    return ','.join(dn.split(',')[1:])

def add_operation(fqdn: str, dn: str,  attributes: dict):
    attr_xml = ""

    parent_dn = get_parent_dn(dn)
    rdn = get_rdn(dn)
    if parent_dn:
        attr_xml += ATTRIBUTE_TYPE_AND_VALUE_FSTRING.format(attribute="ad:container-hierarchy-parent", value=parent_dn)

    if rdn:
        attr_xml += ATTRIBUTE_TYPE_AND_VALUE_FSTRING.format(attribute="ad:relativeDistinguishedName", value=rdn)
    
    for pos, attribute in enumerate(attributes):
        if isinstance(attributes[attribute], SEQUENCE_TYPES):
            for index, value in enumerate(attributes[attribute]):
                attr_xml += ATTRIBUTE_TYPE_AND_VALUE_FSTRING.format(attribute=f"addata:{attribute}", value=value)
        else:
            attr_xml += ATTRIBUTE_TYPE_AND_VALUE_FSTRING.format(attribute=f"addata:{attribute}", value=attributes[attribute])
    add_vars = {
        "fqdn": fqdn,
        "uuid": str(uuid4()),
        "attributes": attr_xml.strip(),
    }
    request = LDAP_ADD_FSTRING.format(**add_vars)
    return request

def add_response_to_dict(response):
    pass


#!/usr/bin/env python3

from .core import (
	LdapParserException, Operators, LdapToken, 
	AttributeParser, DNParser, FilterParser,
	WILDCARD, ATTRIBUTE_OID, EXCEPTION_ATTRIBUTES
)

from .utils import in_exception, LdapObfuscate

from .strategies import (
	oid_attribute_obfuscation, anr_attribute_obfuscation, random_casing_obfuscation,
	prepend_zeros_obfuscation, hex_value_obfuscation, spacing_obfuscation,
	equality_to_approximation_obfuscation, wildcard_expansion_obfuscation
)

class LdapParser(FilterParser):
	"""
	Extended LDAP Filter Parser with modular obfuscation strategies.
	Maintains backward compatibility while providing new modular approach.
	"""
	
	def random_casing(self, parsed_structure=None, probability=0.7):
		"""Random case obfuscation with type awareness"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		random_casing_obfuscation(parsed_structure, self._apply_to_filter_leaves, probability)

	def prepend_zeros_obfuscation(self, parsed_structure=None, max_zeros=5):
		"""Prepend zeros to numeric values and SID components"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		prepend_zeros_obfuscation(parsed_structure, self._apply_to_filter_leaves, max_zeros)

	def hex_value_obfuscation(self, parsed_structure=None, probability=0.5):
		"""Hex encode DN string values"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		hex_value_obfuscation(parsed_structure, self._apply_to_filter_leaves, probability)

	def spacing_obfuscation(self, parsed_structure=None, max_spaces=3):
		"""Add context-aware spacing to attribute values"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		spacing_obfuscation(parsed_structure, self._apply_to_filter_leaves, max_spaces)

	def oid_attribute_obfuscation(self, parsed_structure=None, max_spaces=3, max_zeros=5, include_prefix=True):
		"""Convert attributes to OIDs with spacing and zero padding"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		oid_attribute_obfuscation(parsed_structure, self._apply_to_filter_leaves, max_spaces, max_zeros, include_prefix)

	def anr_attribute_obfuscation(self, parsed_structure=None, anr_set=None):
		"""Convert compatible attributes to ANR"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		anr_attribute_obfuscation(parsed_structure, self._apply_to_filter_leaves, anr_set)

	def equality_to_approximation_obfuscation(self, parsed_structure=None):
		"""Convert equality matches to approximation matches"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		equality_to_approximation_obfuscation(parsed_structure)

	def wildcard_expansion_obfuscation(self, parsed_structure=None):
		"""Safe wildcard expansion for broader matching"""
		if parsed_structure is None:
			parsed_structure = self.parsed_structure
		wildcard_expansion_obfuscation(parsed_structure, self._apply_to_filter_leaves)

	# Legacy method names for backwards compatibility
	def prepend_zeros(self, parsed_structure=None):
		"""Legacy wrapper for prepend_zeros_obfuscation"""
		self.prepend_zeros_obfuscation(parsed_structure)

	def random_hex(self, parsed_structure=None):
		"""Legacy wrapper for hex_value_obfuscation"""
		self.hex_value_obfuscation(parsed_structure)

	def random_wildcards(self, parsed_structure=None):
		"""Legacy wrapper for wildcard_expansion_obfuscation"""
		self.wildcard_expansion_obfuscation(parsed_structure)

	def randomize_oid(self, parsed_structure=None):
		"""Legacy wrapper for oid_attribute_obfuscation"""
		self.oid_attribute_obfuscation(parsed_structure)

	def comparison_operator_obfuscation(self, parsed_structure=None):
		"""Legacy wrapper for equality_to_approximation_obfuscation"""
		self.equality_to_approximation_obfuscation(parsed_structure)

	def boolean_operator_obfuscation(self, parsed_structure=None):
		"""
		Boolean operator obfuscation.
		"""
		return NotImplementedError("Boolean operator obfuscation not yet implemented")

	def append_garbage(self, parsed_structure=None):
		"""
		Append garbage method.
		"""
		return NotImplementedError("Append garbage method not yet implemented")

FilterParser = LdapParser

__all__ = [
	'FilterParser', 'LdapParser', 'DNParser', 'AttributeParser', 'LdapObfuscate',
	'LdapParserException', 'Operators', 'LdapToken', 'in_exception'
] 


from ..templates import LDAP_DELETE_FSTRING

from uuid import uuid4

def delete_operation(fqdn, dn):
    delete_vars = {
        "fqdn": fqdn,
        "uuid": str(uuid4()),
        "object_ref": dn,
    }
    request = LDAP_DELETE_FSTRING.format(**delete_vars)
    return request

def delete_response_to_dict(response):
    pass



from .. import OPERATIONAL_ATTRIBUTES, RESOURCE, RESOURCE_FACTORY, ENUMERATION, ACCOUNT_MANAGEMENT, TOPOLOGY_MANAGEMENT, COMMON_ATTRIBUTES
from ..operation.search import search_operation, handle_str_to_xml, handle_enum_ctx, parse_adws_pull_response, xml_to_dict
from ..operation.modify import modify_operation
from ..operation.delete import delete_operation
from ..operation.add import add_operation
from ..operation.modifyDN import modify_dn_operation
from ..nns import NNS
from ..nmf import NMFConnection
from ..templates import NAMESPACES
from ..error import ADWSError

from ldap3.utils.dn import safe_dn
from ldap3 import (
    ALL_ATTRIBUTES, NO_ATTRIBUTES, ALL_OPERATIONAL_ATTRIBUTES, NTLM, BASE, LEVEL, SUBTREE, 
    STRING_TYPES, get_config_parameter, SEQUENCE_TYPES, MODIFY_ADD, MODIFY_DELETE, 
    MODIFY_REPLACE, MODIFY_INCREMENT, SYNC, SAFE_SYNC, SAFE_RESTARTABLE, ASYNC, LDIF, 
    RESTARTABLE, REUSABLE, MOCK_SYNC, MOCK_ASYNC, ASYNC_STREAM
)
from ldap3.core.connection import CLIENT_STRATEGIES
from ldap3.protocol.rfc2696 import paged_search_control
from ldap3.core.exceptions import LDAPChangeError, LDAPAttributeError, LDAPUnknownStrategyError, LDAPObjectClassError
from ldap3.protocol.formatters.standard import format_attribute_values
from ldap3.strategy.sync import SyncStrategy
from ldap3.strategy.safeSync import SafeSyncStrategy
from ldap3.strategy.safeRestartable import SafeRestartableStrategy
from ldap3.strategy.mockAsync import MockAsyncStrategy
from ldap3.strategy.asynchronous import AsyncStrategy
from ldap3.strategy.reusable import ReusableStrategy
from ldap3.strategy.restartable import RestartableStrategy
from ldap3.strategy.ldifProducer import LdifProducerStrategy
from ldap3.strategy.mockSync import MockSyncStrategy
from ldap3.strategy.asyncStream import AsyncStreamStrategy
from ldap3.utils.conv import to_unicode

import logging
import socket
from copy import deepcopy
from functools import reduce

class Connection(object):
    def __init__(self, server, user, password, domain, lmhash, nthash, raise_exceptions=False, authentication=NTLM, check_names=True, auto_encode=True, client_strategy=SYNC):
        self.server = server
        self.host = server.host
        self.port = server.port
        self.resource = server.resource
        self.user = user
        self.password = password
        self.domain = domain
        self.lmhash = lmhash
        self.nthash = nthash
        self.raise_exceptions = raise_exceptions
        self.authentication = authentication
        self.check_names = check_names
        self.auto_encode = auto_encode
        self.result = {}
        
        # Add ldap3-compatible attributes
        self.bound = False
        self.closed = True
        self.last_error = None
        
        # Add compatibility attributes for connection pooling
        self.nmf = None

        conf_default_pool_name = get_config_parameter('DEFAULT_THREADED_POOL_NAME')
        if client_strategy not in CLIENT_STRATEGIES:
            if self.raise_exceptions:
                raise LDAPUnknownStrategyError('unknown client connection strategy')
            else:
                logging.error('unknown client connection strategy')
                return
        
        self.strategy_type = client_strategy
        if self.strategy_type == SYNC:
            self.strategy = SyncStrategy(self)
        elif self.strategy_type == SAFE_SYNC:
            self.strategy = SafeSyncStrategy(self)
        elif self.strategy_type == SAFE_RESTARTABLE:
            self.strategy = SafeRestartableStrategy(self)
        elif self.strategy_type == ASYNC:
            self.strategy = AsyncStrategy(self)
        elif self.strategy_type == LDIF:
            self.strategy = LdifProducerStrategy(self)
        elif self.strategy_type == RESTARTABLE:
            self.strategy = RestartableStrategy(self)
        elif self.strategy_type == REUSABLE:
            self.strategy = ReusableStrategy(self)
            self.lazy = False
        elif self.strategy_type == MOCK_SYNC:
            self.strategy = MockSyncStrategy(self)
        elif self.strategy_type == MOCK_ASYNC:
            self.strategy = MockAsyncStrategy(self)
        elif self.strategy_type == ASYNC_STREAM:
            self.strategy = AsyncStreamStrategy(self)
        else:
            if self.raise_exceptions:
                raise LDAPUnknownStrategyError('unknown strategy')
            else:
                logging.error('unknown strategy')
                
    def unbind(self):
        """
        Close the connection and mark as unbound.
        Compatible with ldap3.Connection.unbind()
        """
        try:
            if hasattr(self, 'nmf') and self.nmf is not None:
                try:
                    self.nmf._sock.close()
                except Exception:
                    pass
                self.nmf = None
        except Exception:
            pass
        finally:
            self.bound = False
            self.closed = True
    
    def close(self):
        """
        Close the connection. Alias for unbind() for compatibility.
        """
        self.unbind()
    
    def is_connection_alive(self):
        """
        Check if the ADWS connection is alive.
        Compatible with the connection pool interface.
        """
        try:
            # For ADWS, if we're marked as bound and not closed, consider it alive
            # The socket-level checks might be too strict for ADWS connections
            if self.closed or not self.bound:
                return False
            
            # Basic check that we have an nmf connection
            if not hasattr(self, 'nmf') or self.nmf is None:
                return False
            
            # For now, return True if basic state is valid
            # ADWS connections are more complex and may not follow standard socket patterns
            return True
        except Exception:
            return False
    
    def abandon(self, message_id):
        """
        ADWS abandon operation (lightweight keep-alive).
        Compatible with ldap3 abandon for keep-alive purposes.
        """
        # For ADWS, we don't have a direct abandon operation
        # This is used primarily for keep-alive testing
        return self.is_connection_alive()
    
    def keep_alive(self):
        """
        Perform a lightweight ADWS operation to keep the connection alive.
        Compatible with the connection pool interface.
        
        Returns:
            bool: True if connection is still alive, False otherwise
        """
        try:
            # Use the abandon method for lightweight keep-alive
            result = self.abandon(0)
            if result:
                logging.debug(f"[ADWS Connection] Keep-alive check: Success")
            return result
        except Exception as e:
            logging.debug(f"[ADWS Connection] Keep-alive check failed: {str(e)}")
            return False

    def _create_NNS_from_auth(self, sock: socket.socket) -> NNS:
        if self.authentication == NTLM:
            return NNS(
                socket=sock,
                fqdn=self.host,
                domain=self.domain,
                username=self.user,
                password=self.password,
                nt=self.nthash if self.nthash else "",
                lm=self.lmhash if self.lmhash else ""
            )
        raise NotImplementedError

    def _prepare_return_value(self, response):
        error_detail = response.get("ErrorDetail", {})
        fault_detail = error_detail.get("FaultDetail", {})
        specific_error = {}

        if isinstance(fault_detail, dict) and fault_detail:
            specific_error = next(iter(fault_detail.values()), {})
            if not isinstance(specific_error, dict):
                specific_error = {}

        self.result['message'] = specific_error.get("ExtendedErrorMessage", fault_detail.get("ExtendedErrorMessage", response.get("Error", "Unknown error")))
        self.result['description'] = specific_error.get("Message", fault_detail.get("ExtendedErrorDescription", response.get("Error", "Unknown error")))

        error_code_val = specific_error.get("ErrorCode", fault_detail.get("ErrorCode", 0))
        try:
            self.result['result'] = int(error_code_val)
        except (ValueError, TypeError):
            self.result['result'] = str(error_code_val) if error_code_val else 0

        win32_error_code_val = specific_error.get("Win32ErrorCode", 0)
        try:
            self.result['win32_error_code'] = int(win32_error_code_val)
        except (ValueError, TypeError):
            self.result['win32_error_code'] = 0

        self.result['short_message'] = specific_error.get("ShortMessage", fault_detail.get("ShortError", "Unknown"))

        self.entries = response.get("entries", [])
        response['entries'] = self.entries

    def refresh_server_info(self):
        self.server.get_info_from_server(self)

    def send_and_recv(self, request):
        self.nmf.send(request)
        response = self.nmf.recv()
        return response

    def search(self,
               search_base,
               search_filter,
               search_scope=SUBTREE,
               attributes=None,
               size_limit=0,
               time_limit=0,
               types_only=False,
               get_operational_attributes=False,
               controls=None,
               paged_size=None,
               paged_criticality=False,
               paged_cookie=None,
               auto_escape=None):
        """
        Perform a search operation on the server.
        """
        if self.resource != ENUMERATION:
            logging.debug("Search is only supported for the \"Enumeration\" resource, reconnecting to the server")
            self.reconnect(ENUMERATION)

        conf_attributes_excluded_from_check = [v.lower() for v in get_config_parameter('ATTRIBUTES_EXCLUDED_FROM_CHECK')]

        if self.check_names and search_base:
            search_base = safe_dn(search_base)

        if not attributes:
            attributes = [NO_ATTRIBUTES]
        elif attributes == ALL_ATTRIBUTES:
            attributes = [ALL_ATTRIBUTES]

        if isinstance(attributes, STRING_TYPES):
            attributes = [attributes]

        if get_operational_attributes and isinstance(attributes, list):
            attributes.append(ALL_OPERATIONAL_ATTRIBUTES)
        elif get_operational_attributes and isinstance(attributes, tuple):
            attributes += (ALL_OPERATIONAL_ATTRIBUTES, )  # concatenate tuple

        if isinstance(paged_size, int):
            if controls is None:
                controls = []
            else:
                controls = list(controls)
            controls.append(paged_search_control(paged_criticality, paged_size, paged_cookie))

        if self.server and self.server.schema and self.check_names:
            for attribute_name in attributes:
                if ';' in attribute_name:  # remove tags
                    attribute_name_to_check = attribute_name.split(';')[0]
                else:
                    attribute_name_to_check = attribute_name
                if (
                    self.server.schema
                    and attribute_name_to_check.lower() not in conf_attributes_excluded_from_check
                    and attribute_name_to_check not in self.server.schema.attribute_types
                    and attribute_name_to_check not in OPERATIONAL_ATTRIBUTES
                ):
                    self.last_error = 'invalid attribute type ' + attribute_name_to_check
                    if self.raise_exceptions:
                        raise LDAPAttributeError(self.last_error)

        try:
            # make search operation to the server before pulling the results
            request = search_operation(self.host, search_base, search_filter, search_scope, attributes)
            response = self.send_and_recv(request)
            et = handle_str_to_xml(response)
            if not et:
                raise ValueError("was unable to parse xml from the server response")

            enum_ctx = et.find(".//wsen:EnumerationContext", NAMESPACES).text
            if enum_ctx is None:
                raise ValueError("Enum Context not found in response")

            # now we need to pull the results from the server
            pull_request = handle_enum_ctx(self.host, enum_ctx)
            response = self.send_and_recv(pull_request)
            #results = parse_adws_pull_response(response)
            resp_dict = xml_to_dict(response, attributes)
            if resp_dict.get('entries'):
                for entry in resp_dict['entries']:
                    for attribute in entry['attributes']:
                        entry['attributes'][attribute] = format_attribute_values(self.server.schema, attribute, entry['attributes'][attribute], self.server.custom_formatter)
            #self._prepare_return_value(resp_dict)
            return resp_dict.get('entries', [])
        except ADWSError as e:
            if "size limit was exceeded" in str(e).lower() and attributes and isinstance(attributes, list):
                logging.warning("Size limit was exceeded, trying default attributes")
                attributes = COMMON_ATTRIBUTES
                return self.search(search_base, search_filter, search_scope, attributes, size_limit, time_limit, types_only, get_operational_attributes, controls, paged_size, paged_criticality, paged_cookie, auto_escape)
            else:
                raise

    def modify(self,
               dn,
               changes,
               controls=None):
        """
        Modify attributes of entry

        - changes is a dictionary in the form {'attribute1': change), 'attribute2': [change, change, ...], ...}
        - change is (operation, [value1, value2, ...])
        - operation is 0 (MODIFY_ADD), 1 (MODIFY_DELETE), 2 (MODIFY_REPLACE), 3 (MODIFY_INCREMENT)
        """
        if self.resource != RESOURCE:
            logging.debug("Modify is only supported for the \"Resource\" resource, reconnecting to the server")
            self.reconnect(RESOURCE)

        conf_attributes_excluded_from_check = [v.lower() for v in get_config_parameter('ATTRIBUTES_EXCLUDED_FROM_CHECK')]
        
        if self.check_names:
            dn = safe_dn(dn)

        if not isinstance(changes, dict):
                self.last_error = 'changes must be a dictionary'
                raise LDAPChangeError(self.last_error)

        if not changes:
            self.last_error = 'no changes in modify request'
            raise LDAPChangeError(self.last_error)

        changelist = dict()
        for attribute_name in changes:
            if self.server and self.server.schema and self.check_names:
                if ';' in attribute_name:  # remove tags for checking
                    attribute_name_to_check = attribute_name.split(';')[0]
                else:
                    attribute_name_to_check = attribute_name

                if self.server.schema.attribute_types and attribute_name_to_check.lower() not in conf_attributes_excluded_from_check and attribute_name_to_check not in self.server.schema.attribute_types:
                    self.last_error = 'invalid attribute type ' + attribute_name_to_check
                    raise LDAPAttributeError(self.last_error)
        change = changes[attribute_name]
        if isinstance(change, SEQUENCE_TYPES) and change[0] in [MODIFY_ADD, MODIFY_DELETE, MODIFY_REPLACE, MODIFY_INCREMENT, 0, 1, 2, 3]:
            if len(change) != 2:
                self.last_error = 'malformed change'
                raise LDAPChangeError(self.last_error)
            changelist[attribute_name] = [change]  # insert change in a list
        else:
            for change_operation in change:
                if len(change_operation) != 2 or change_operation[0] not in [MODIFY_ADD, MODIFY_DELETE, MODIFY_REPLACE, MODIFY_INCREMENT, 0, 1, 2, 3]:
                    self.last_error = 'invalid change list'
                    raise LDAPChangeError(self.last_error)
            changelist[attribute_name] = change
        
        request = modify_operation(self.host, dn, changelist, self.auto_encode, self.server.schema if self.server else None, validator=self.server.custom_validator if self.server else None, check_names=self.check_names)
        response = self.send_and_recv(request)
        resp_dict = xml_to_dict(response)
        self._prepare_return_value(resp_dict)
        if resp_dict.get("Error"):
            if self.raise_exceptions:
                raise ADWSError(f"{self.result['short_message']}: {self.result['description']}")
            else:
                logging.error(f"{self.result['short_message']}: {self.result['description']}")
                return False
        else:
            return True

    def modify_dn(self,
                  dn,
                  relative_dn,
                  delete_old_dn=True,
                  new_superior=None,
                  controls=None):
        """
        Modify DN of the entry or performs a move of the entry in the
        DIT.
        """

        if self.resource != RESOURCE:
            logging.debug("ModifyDN is only supported for the \"Resource\" resource, reconnecting to the server")
            self.reconnect(RESOURCE)

        request = modify_dn_operation(self.host, dn, relative_dn, delete_old_dn, new_superior)
        response = self.send_and_recv(request)
        resp_dict = xml_to_dict(response)
        self._prepare_return_value(resp_dict)
        if resp_dict.get("Error"):
            if self.raise_exceptions:
                raise ADWSError(f"{self.result['short_message']}: {self.result['description']}")
            else:
                logging.error(f"{self.result['short_message']}: {self.result['description']}")
                return False
        else:
            return True

    def add(self,
            dn,
            object_class=None,
            attributes=None,
            controls=None):
        """
        Add dn to the DIT, object_class is None, a class name or a list
        of class names.

        Attributes is a dictionary in the form 'attr': 'val' or 'attr':
        ['val1', 'val2', ...] for multivalued attributes
        """
        if self.resource != RESOURCE_FACTORY:
            logging.debug("Add is only supported for the \"ResourceFactory\" resource, reconnecting to the server")
            self.reconnect(RESOURCE_FACTORY)

        conf_attributes_excluded_from_check = [v.lower() for v in get_config_parameter('ATTRIBUTES_EXCLUDED_FROM_CHECK')]
        conf_classes_excluded_from_check = [v.lower() for v in get_config_parameter('CLASSES_EXCLUDED_FROM_CHECK')]

        self.last_error = None
        _attributes = deepcopy(attributes)
        if self.check_names:
            dn = safe_dn(dn)

        attr_object_class = []
        if object_class is None:
            parm_object_class = []
        else:
            parm_object_class = list(object_class) if isinstance(object_class, SEQUENCE_TYPES) else [object_class]

        object_class_attr_name = ''
        if _attributes:
            for attr in _attributes:
                if attr.lower() == 'objectclass':
                    object_class_attr_name = attr
                    attr_object_class = list(_attributes[object_class_attr_name]) if isinstance(_attributes[object_class_attr_name], SEQUENCE_TYPES) else [_attributes[object_class_attr_name]]
                    break
        else:
            _attributes = dict()

        if not object_class_attr_name:
                object_class_attr_name = 'objectClass'

        attr_object_class = [to_unicode(object_class) for object_class in attr_object_class]  # converts objectclass to unicode in case of bytes value
        _attributes[object_class_attr_name] = reduce(lambda x, y: x + [y] if y not in x else x, parm_object_class + attr_object_class, [])  # remove duplicate ObjectClasses

        if not _attributes[object_class_attr_name]:
            raise LDAPObjectClassError('objectClass attribute is mandatory')

        if self.server and self.server.schema and self.check_names:
            for object_class_name in _attributes[object_class_attr_name]:
                if object_class_name.lower() not in conf_classes_excluded_from_check and object_class_name not in self.server.schema.object_classes:
                    raise LDAPObjectClassError('invalid object class ' + str(object_class_name))

            for attribute_name in _attributes:
                if ';' in attribute_name:  # remove tags for checking
                    attribute_name_to_check = attribute_name.split(';')[0]
                else:
                    attribute_name_to_check = attribute_name

                if attribute_name_to_check.lower() not in conf_attributes_excluded_from_check and attribute_name_to_check not in self.server.schema.attribute_types:
                    raise LDAPAttributeError('invalid attribute type ' + attribute_name_to_check)

        request = add_operation(self.host, dn, _attributes)
        response = self.send_and_recv(request)
        resp_dict = xml_to_dict(response)
        self._prepare_return_value(resp_dict)
        if resp_dict.get("Error"):
            if self.raise_exceptions:
                raise ADWSError(f"{self.result['short_message']}: {self.result['description']}")
            else:
                logging.error(f"{self.result['short_message']}: {self.result['description']}")
                return False
        else:
            return True

    def delete(self,
               dn,
               controls=None):
        """
        Delete the entry identified by the DN from the DIB.
        """
        if self.resource != RESOURCE:
            logging.debug("Delete is only supported for the \"Resource\" resource, reconnecting to the server")
            self.reconnect(RESOURCE)

        request = delete_operation(self.host, dn)
        response = self.send_and_recv(request)
        resp_dict = xml_to_dict(response)
        self._prepare_return_value(resp_dict)
        if resp_dict.get("Error"):
            if self.raise_exceptions:
                raise ADWSError(self.result['message'])
            else:
                logging.error(self.result['message'])
                return False
        else:
            return True

    def connect(self, resource=None, get_info=False):
        """Connect to the specified ADWS endpoint at the
        Args:
            resource (str): endpoint to connect to <'Resource', 'ResourceFactory',
                'Enumeration', AccountManagement',  'TopologyManagement'>
        """
        try:
            server_address: tuple[str, int] = (self.host, self.port)
            resource = resource if resource is not None else self.resource
            logging.debug(f"Connecting to {self.host} for {resource}")
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(server_address)
            self.nmf = NMFConnection(
                self._create_NNS_from_auth(sock),
                fqdn=self.host,
                port=self.port
            )
            self.nmf.connect(f"Windows/{resource}")
            self.resource = resource
            
            # Set connection state
            self.bound = True
            self.closed = False
            
            if get_info:
                self.refresh_server_info()
            return self.server, self
        except Exception as e:
            # Connection failed
            self.bound = False
            self.closed = True
            raise ADWSError(f"Failed to connect: {str(e)}")

    def reconnect(self, resource=None):
        """
        Reconnect to the server with a new resource
        """
        try:
            logging.debug(f"Reconnecting to {self.host} for {resource}")
            if hasattr(self, 'nmf') and self.nmf is not None:
                try:
                    self.nmf._sock.close()
                except Exception:
                    pass
            return self.connect(resource)
        except Exception as e:
            self.bound = False
            self.closed = True
            raise ADWSError(f"Failed to reconnect: {str(e)}")



#!/usr/bin/env python3

import logging
import json
from typing import Any, Optional
import datetime
import os
from powerview.modules.smbclient import SMBClient
from powerview.utils.helpers import is_ipaddress, is_valid_fqdn, host2ip

def _format_mcp_response(
	data: Any = None,
	error: Optional[str] = None,
	success: Optional[bool] = None,
	message: Optional[str] = None,
	status_override: Optional[str] = None
) -> str:
	"""Formats the response payload into a standardized JSON string for MCP."""
	response = {}
	status = "unknown"

	if error:
		status = "error"
		response["error"] = str(error)
	elif success is True:
		status = "success"
		response["message"] = message or "Operation successful."
		if data is not None:
			response["data"] = data
	elif success is False:
		status = "failure"
		response["error"] = message or "Operation failed."
	elif data is not None:
		if isinstance(data, (list, dict)) and not data:
			status = status_override or "not_found"
			response["message"] = message or "No results found matching criteria."
		else:
			status = status_override or "success"
			response["data"] = data
	elif status_override:
		status = status_override
		response["message"] = message or f"Status: {status_override}"
	else:
		status = "no_content"
		response["message"] = message or "Operation completed with no specific content returned."

	response["status"] = status

	try:
		return json.dumps(response, default=str)
	except TypeError as e:
		logging.error(f"Failed to serialize MCP response: {e}. Response data: {response}")
		return json.dumps({"status": "error", "error": f"Internal serialization error: {e}"})

def setup_tools(mcp, powerview_instance):
	"""Register all PowerView tools with the MCP server."""

	@mcp.tool()
	async def login_as(
		username: str,
		password: str | None = None,
		domain: str | None = None,
		lmhash: str | None = None,
		nthash: str | None = None,
		auth_aes_key: str | None = None
	) -> str:
		"""
		Login as a different user from the current context.
		
		Args:
			username: The username to login as.
			password: The password to login with. Mutually exclusive with lmhash and nthash and auth_aes_key.
			domain: The domain to login to. Optional, will use current domain if not provided.
			lmhash: The LM hash to use for the login. Mutually exclusive with password.
			nthash: The NTHash to use for the login. Mutually exclusive with password. If 32 hash length, will be considered as NTHash. Leave blank for lmhash and password.
			auth_aes_key: The AES key to use for the login. Mutually exclusive with password. If longer than 32 characters, will be considered as auth_aes_key. Leave blank for password.
		"""
		try:
			success = powerview_instance.login_as(
				username=username,
				password=password,
				domain=domain,
				lmhash=lmhash,
				nthash=nthash,
				auth_aes_key=auth_aes_key
			)

			if success:
				current_identity = powerview_instance.conn.who_am_i()
				message = f"Successfully logged in as {current_identity}"
				return _format_mcp_response(success=True, message=message)
			else:
				message = f"Failed to login as {username}@{domain or powerview_instance.domain}. Check credentials or permissions."
				logging.warning(message)
				return _format_mcp_response(success=False, message=message)

		except Exception as e:
			logging.error(f"Unexpected exception during login_as for {username}: {str(e)}")
			return _format_mcp_response(error=f"An unexpected error occurred during login: {str(e)}")

	@mcp.tool()
	async def get_domain_user(
		identity: str = "*",
		properties: str = "*",
		preauthnotrequired: bool = False,
		passnotrequired: bool = False,
		password_expired: bool = False,
		admincount: bool = False,
		trustedtoauth: bool = False,
		allowdelegation: bool = False,
		disallowdelegation: bool = False,
		rbcd: bool = False,
		unconstrained: bool = False,
		shadowcred: bool = False,
		spn: bool = False,
		enabled: bool = False,
		disabled: bool = False,
		lockout: bool = False,
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get information about domain users with comprehensive filtering options.

		Args:
			identity: Filter by user identity (sAMAccountName, SID, GUID, DN). Defaults to '*' (all users).
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			preauthnotrequired: Filter for users with 'DONT_REQ_PREAUTH' set.
			passnotrequired: Filter for users with 'PASSWORD_NOT_REQUIRED' set.
			password_expired: Filter for users whose password has expired.
			admincount: Filter for users with adminCount=1.
			trustedtoauth: Filter for users/computers trusted for constrained delegation.
			allowdelegation: Filter for accounts allowed to delegate.
			disallowdelegation: Filter for accounts explicitly disallowed from delegation.
			rbcd: Filter for accounts with Resource-Based Constrained Delegation configured.
			unconstrained: Filter for accounts configured for unconstrained delegation.
			shadowcred: Filter for accounts vulnerable to Shadow Credentials attack (msDS-KeyCredentialLink).
			spn: Filter for accounts with a Service Principal Name.
			enabled: Filter for enabled accounts.
			disabled: Filter for disabled accounts.
			lockout: Filter for locked out accounts.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'preauthnotrequired': preauthnotrequired,
				'passnotrequired': passnotrequired,
				'password_expired': password_expired,
				'admincount': admincount,
				'trustedtoauth': trustedtoauth,
				'allowdelegation': allowdelegation,
				'disallowdelegation': disallowdelegation,
				'rbcd': rbcd,
				'unconstrained': unconstrained,
				'shadowcred': shadowcred,
				'spn': spn,
				'enabled': enabled,
				'disabled': disabled,
				'lockout': lockout,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainUser'
			})
			result = powerview_instance.get_domainuser(args=args)
			return _format_mcp_response(data=result, message="No users found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_user: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_computer(
		identity: str = "*",
		properties: str = "*",
		enabled: bool = False,
		disabled: bool = False,
		unconstrained: bool = False,
		trustedtoauth: bool = False,
		allowdelegation: bool = False,
		disallowdelegation: bool = False,
		rbcd: bool = False,
		shadowcred: bool = False,
		spn: bool = False,
		printers: bool = False,
		ping: bool = False,
		resolveip: bool = False,
		resolvesids: bool = False,
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False,
		laps: bool = False,
		bitlocker: bool = False,
		gmsapassword: bool = False,
		pre2k: bool = False,
		excludedcs: bool = False
	) -> str:
		"""Get information about domain computers with comprehensive filtering options.

		Args:
			identity: Filter by computer identity (sAMAccountName, DNS Hostname, SID, GUID, DN). Defaults to '*'.
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			enabled: Filter for enabled computer accounts.
			disabled: Filter for disabled computer accounts.
			unconstrained: Filter for computers configured for unconstrained delegation.
			trustedtoauth: Filter for computers trusted for constrained delegation.
			allowdelegation: Filter for computers allowed to delegate. (Note: Often used with users)
			disallowdelegation: Filter for computers explicitly disallowed from delegation. (Note: Often used with users)
			rbcd: Filter for computers with Resource-Based Constrained Delegation configured.
			shadowcred: Filter for computers vulnerable to Shadow Credentials attack (msDS-KeyCredentialLink).
			spn: Filter for computers with a Service Principal Name.
			printers: Filter for computers that are print servers (may involve heuristics or specific SPNs).
			ping: Attempt to ping the computer (Note: MCP implementation might not fully support this; relies on underlying library).
			resolveip: Attempt to resolve the IP address for the computer.
			resolvesids: Resolve SIDs found in properties like PrimaryGroupID.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
			laps: Filter for computers potentially managed by LAPS.
			bitlocker: Filter for computers with BitLocker recovery information.
			gmsapassword: Filter for computers associated with GMSA passwords.
			pre2k: Filter for pre-Windows 2000 compatible computers.
			excludedcs: Exclude Domain Controllers from the results.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'enabled': enabled,
				'disabled': disabled,
				'unconstrained': unconstrained,
				'trustedtoauth': trustedtoauth,
				'allowdelegation': allowdelegation,
				'disallowdelegation': disallowdelegation,
				'rbcd': rbcd,
				'shadowcred': shadowcred,
				'spn': spn,
				'printers': printers,
				'ping': ping, # Pass along, PowerView method needs to handle it
				'resolveip': resolveip,
				'resolvesids': resolvesids,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'laps': laps,
				'bitlocker': bitlocker,
				'gmsapassword': gmsapassword,
				'pre2k': pre2k,
				'excludedcs': excludedcs,
				'module': 'Get-DomainComputer'
			})
			result = powerview_instance.get_domaincomputer(args=args)
			return _format_mcp_response(data=result, message="No computers found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_computer: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_group(
		identity: str = "*",
		properties: str = "*",
		admincount: bool = False,
		ldapfilter: str = "",
		memberidentity: str = "",
		no_cache: bool = False,
		searchbase: str = "",
		raw: bool = False,
		no_vuln_check: bool = False
	) -> str:
		"""Get information about domain groups.

		Args:
			identity: Filter by group identity (sAMAccountName, SID, GUID, DN). Defaults to '*'.
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			admincount: Filter for groups with adminCount=1 (protected groups).
			ldapfilter: Custom LDAP filter string.
			memberidentity: Find groups where this identity (user/group DN or sAMAccountName) is a member.
			no_cache: Bypass the cache and perform a live query.
			searchbase: Specify the search base DN.
			raw: Return raw LDAP entries without formatting.
			no_vuln_check: Disable vulnerability checks.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'admincount': admincount,
				'ldapfilter': ldapfilter,
				'memberidentity': memberidentity,
				'no_cache': no_cache,
				'searchbase': searchbase if searchbase else None,
				'raw': raw,
				'no_vuln_check': no_vuln_check,
				'module': 'Get-DomainGroup'
			})
			result = powerview_instance.get_domaingroup(args=args)
			return _format_mcp_response(data=result, message="No groups found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_group: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_group_member(
		identity: str,
		multiple: bool = False,
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False,
		ldapfilter: str = ""
	) -> str:
		"""Get members of a domain group.

		Args:
			identity: Identity of the group (sAMAccountName, SID, GUID, DN) to get members from.
			multiple: If specified, handle multiple groups matching the identity (Not typically needed if identity is specific).
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries for members without formatting.
			ldapfilter: Custom LDAP filter string to apply to the group search itself.
		"""
		try:
			args = type('Args', (), {
				'identity': identity,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'ldapfilter': ldapfilter,
				'module': 'Get-DomainGroupMember'
			})
			result = powerview_instance.get_domaingroupmember(identity=identity, multiple=multiple, args=args)
			return _format_mcp_response(data=result, message=f"No members found for group '{identity}'")
		except Exception as e:
			logging.error(f"Error in get_domain_group_member: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_controller(
		identity: str = "*",
		properties: str = "*",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False,
		resolvesids: bool = False
	) -> str:
		"""Get information about domain controllers.

		Args:
			identity: Filter by DC identity (hostname, DN). Defaults to '*'.
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
			resolvesids: Resolve SIDs found in properties.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'resolvesids': resolvesids,
				'module': 'Get-DomainController'
			})
			result = powerview_instance.get_domaincontroller(args=args)
			return _format_mcp_response(data=result, message="No domain controllers found")
		except Exception as e:
			logging.error(f"Error in get_domain_controller: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_trust(
		identity: str = "*",
		properties: str = "*",
		searchbase: str = "",
		ldapfilter: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False,
		sd_flag: str = "" # Note: sd_flag is not in parser, might be internal
	) -> str:
		"""Get information about domain trusts.

		Args:
			identity: Filter by trust identity (target domain name, DN). Defaults to '*'.
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			searchbase: Specify the search base DN (usually the domain root or Configuration NC).
			ldapfilter: Custom LDAP filter string.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
			sd_flag: Security Descriptor flag (internal use, likely not needed via MCP).
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'searchbase': searchbase if searchbase else None,
				'ldapfilter': ldapfilter,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'sd_flag': sd_flag if sd_flag else None,
				'module': 'Get-DomainTrust'
			})
			result = powerview_instance.get_domaintrust(args=args)
			return _format_mcp_response(data=result, message="No trusts found")
		except Exception as e:
			logging.error(f"Error in get_domain_trust: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain(
		identity: str = "*",
		properties: str = "*",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get domain information.

		Args:
			identity: Filter by domain identity (name, DN). Defaults to '*' or current domain.
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN (usually the domain root).
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-Domain'
			})
			result = powerview_instance.get_domain(args=args)
			return _format_mcp_response(data=result, message="No domain information found")
		except Exception as e:
			logging.error(f"Error in get_domain: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_object_acl(
		identity: str = "*",
		security_identifier: str = "",
		ldapfilter: str = "",
		resolveguids: bool = False,
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get the ACLs for domain objects. IMPORTANT: When analyzing attack paths, always use `security_identifier` to filter ACLs by a specific identity (like your current user) instead of all ACLs. Not doing so will miss critical attack paths.

		Args:
			identity: Identity of the object (DN, sAMAccountName, SID, GUID) to get ACLs for. Defaults to '*'. Leave blank to get ACLs for all objects.
			security_identifier: CRITICAL FOR ATTACK PATH ANALYSIS - Filter ACEs to only show ACLs that the identity has on the object. Useful to find attack paths for a specific identity like your current user. Can be a SID, username, or DN.
			ldapfilter: Custom LDAP filter string (applied to the object search if identity is wildcard).
			resolveguids: Resolve GUIDs in ACEs to friendly names (requires schema access).
			searchbase: Specify the search base DN for the object identity.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			args = type('Args', (), {
				'identity': identity,
				'ldapfilter': ldapfilter,
				'security_identifier': security_identifier,
				'resolveguids': resolveguids,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainObjectAcl'
			})
			result = powerview_instance.get_domainobjectacl(args=args)
			return _format_mcp_response(data=result, message="No ACLs found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_object_acl: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_ou(
		identity: str = "",
		properties: str = "",
		resolve_gplink: bool = False,
		gplink: str = "",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get information about organizational units (OUs).

		Args:
			identity: Filter by OU identity (name, DN). Defaults to empty (likely means all OUs).
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			resolve_gplink: Resolve gpLink attribute GUIDs to GPO display names.
			gplink: Filter OUs by a specific GPO GUID linked to them.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'resolve_gplink': resolve_gplink,
				'gplink': gplink if gplink else None,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainOU'
			})
			result = powerview_instance.get_domainou(args=args)
			return _format_mcp_response(data=result, message="No OUs found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_ou: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_gpo(
		identity: str = "",
		properties: str = "",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get information about Group Policy Objects (GPOs).

		Args:
			identity: Filter by GPO identity (DisplayName, GUID, DN). Defaults to empty (likely means all GPOs).
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN (usually Configuration NC or domain root).
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainGPO'
			})
			result = powerview_instance.get_domaingpo(args=args)
			return _format_mcp_response(data=result, message="No GPOs found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_gpo: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_gpo_localgroup(
		identity: str = "",
	) -> str:
		"""Get local group membership from GPOs."""
		try:
			args = type('Args', (), {
				'identity': identity,
				'module': 'Get-DomainGPOLocalGroup'
			})
			result = powerview_instance.get_domaingpolocalgroup(args=args)
			return _format_mcp_response(data=result, message="No GPO local group info found")
		except Exception as e:
			logging.error(f"Error in get_domain_gpo_localgroup: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_gpo_settings(
		identity: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Parse and return GPO settings from SYSVOL.

		Args:
			identity: Filter by GPO identity (DisplayName, GUID, DN). Required, get gpo name with `get_domain_gpo` tool.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			if not identity:
				return _format_mcp_response(error="Identity is required")

			args = type('Args', (), {
				'identity': identity,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainGPOSettings'
			})
			result = powerview_instance.get_domaingposettings(args=args)
			return _format_mcp_response(data=result, message="No GPO settings found")
		except Exception as e:
			logging.error(f"Error in get_domain_gpo_settings: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_dns_zone(
		identity: str = "",
		properties: str = "",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get information about DNS zones.

		Args:
			identity: Filter by DNS zone identity (name, DN). Defaults to empty (likely means all zones).
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN (usually under System container or application partitions).
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity, # Also passed separately to func
				'properties': props, # Also passed separately to func
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'zonename': None,
				'module': 'Get-DomainDNSZone'
			})
			result = powerview_instance.get_domaindnszone(identity=identity, properties=props, args=args)
			return _format_mcp_response(data=result, message="No DNS zones found")
		except Exception as e:
			logging.error(f"Error in get_domain_dns_zone: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def invoke_kerberoast(
		identity: str = "",
		ldapfilter: str = "",
		searchbase: str = "",
		opsec: bool = False,
		no_cache: bool = False,
	) -> str:
		"""Perform Kerberoasting against service accounts.

		Args:
			identity: Filter for specific user/computer identities to target.
			ldapfilter: Custom LDAP filter to find target accounts (e.g., filter by group).
			searchbase: Specify the search base DN for target accounts.
			opsec: Perform OpSec-safe Kerberoasting (request TGS for krbtgt).
			no_cache: Bypass LDAP query cache when searching for targets.
		"""
		try:
			args = type('Args', (), {
				'identity': identity,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'opsec': opsec,
				'no_cache': no_cache,
				# target_domain is internal detail?
				'module': 'Invoke-Kerberoast'
			})
			result = powerview_instance.invoke_kerberoast(args=args)
			return _format_mcp_response(data=result, message="No kerberoastable accounts found")
		except Exception as e:
			logging.error(f"Error in invoke_kerberoast: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def invoke_asreproast(
		identity: str = "",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
	) -> str:
		"""Perform AS-REP Roasting against accounts with Kerberos pre-authentication disabled.

		Args:
			identity: Filter for specific user identities to target.
			ldapfilter: Custom LDAP filter to find target accounts.
			searchbase: Specify the search base DN for target accounts.
			no_cache: Bypass LDAP query cache when searching for targets.
		"""
		try:
			args = type('Args', (), {
				'identity': identity,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'module': 'Invoke-ASREPRoast'
			})
			result = powerview_instance.invoke_asreproast(args=args)
			return _format_mcp_response(data=result, message="No AS-REP roastable accounts found")
		except Exception as e:
			logging.error(f"Error in invoke_asreproast: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_ca(
		identity: str = "",
		properties: str = "",
		check_all: bool = False,
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get information about domain certificate authorities.

		Args:
			identity: Filter by CA identity (name, DN). Defaults to empty (all CAs).
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			check_all: Check all CA configurations (recommended to get all cert managers and web enrollment status).
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN (usually Configuration NC).
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'check_all': check_all,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainCA'
			})
			result = powerview_instance.get_domainca(args=args)
			return _format_mcp_response(data=result, message="No certificate authorities found")
		except Exception as e:
			logging.error(f"Error in get_domain_ca: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_ca_template(
		identity: str = "",
		properties: str = "",
		vulnerable: bool = False,
		resolve_sids: bool = False,
		enabled: bool = False,
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get information about certificate templates in the domain.

		Args:
			identity: Filter by template identity (DisplayName, Name, OID, DN). Defaults to empty (all templates).
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			vulnerable: Filter for templates potentially vulnerable to abuse (e.g., ESC1, ESC2, etc.).
			resolve_sids: Resolve SIDs in template ACLs/properties to names.
			enabled: Filter for enabled templates.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN (usually Configuration NC).
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'vulnerable': vulnerable,
				'resolve_sids': resolve_sids,
				'enabled': enabled,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainCATemplate'
			})
			result = powerview_instance.get_domaincatemplate(
				args=args
			)
			return _format_mcp_response(data=result, message="No certificate templates found")
		except Exception as e:
			logging.error(f"Error in get_domain_ca_template: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_gmsa(
		identity: str = "",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""Get information about Group Managed Service Accounts (gMSAs).

		Args:
			identity: Filter by gMSA identity (sAMAccountName, SID, DN). Defaults to empty (all gMSAs).
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			args = type('Args', (), {
				'identity': identity,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'module': 'Get-DomainGMSA'
			})
			result = powerview_instance.get_domaingmsa(
				identity=identity if identity else None,
				args=args,
				no_cache=no_cache,
				no_vuln_check=no_vuln_check,
				raw=raw
			)
			return _format_mcp_response(data=result, message="No Group Managed Service Accounts found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_gmsa: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_domain_object_owner(
		identity: str = "*",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""
		Gets the owner of specified domain objects.

		Args:
			identity: Object identity (name, SID, GUID, DN) to find the owner for. Defaults to '*'.
			searchbase: Specify the search base DN.
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			args = type('Args', (), {
				'identity': identity,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'server': None,
				'select': None,
				'where': None,
				'tableview': '',
				'sort_by': None,
				'count': False,
				'outfile': None,
				'nowrap': False,
				'module': 'Get-DomainObjectOwner'
			})
			result = powerview_instance.get_domainobjectowner(args=args)
			return _format_mcp_response(data=result, message="No object owner information found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_domain_object_owner: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_exchange_server(
		identity: str = "*",
		properties: str = "*",
		ldapfilter: str = "",
		searchbase: str = "",
		no_cache: bool = False,
		no_vuln_check: bool = False,
		raw: bool = False
	) -> str:
		"""
		Get information about Exchange servers in the domain.

		Args:
			identity: Filter by Exchange server identity (name, DN). Defaults to '*'.
			properties: Comma-separated list of properties to retrieve. Defaults to '*'.
			ldapfilter: Custom LDAP filter string.
			searchbase: Specify the search base DN (usually Configuration NC).
			no_cache: Bypass the cache and perform a live query.
			no_vuln_check: Disable vulnerability checks.
			raw: Return raw LDAP entries without formatting.
		"""
		try:
			props = properties.split(",") if properties else []
			args = type('Args', (), {
				'identity': identity,
				'properties': props,
				'ldapfilter': ldapfilter,
				'searchbase': searchbase if searchbase else None,
				'no_cache': no_cache,
				'no_vuln_check': no_vuln_check,
				'raw': raw,
				'server': None,
				'select': None,
				'where': None,
				'tableview': '',
				'sort_by': None,
				'count': False,
				'outfile': None,
				'module': 'Get-ExchangeServer'
			})
			result = powerview_instance.get_exchangeserver(args=args)
			return _format_mcp_response(data=result, message="No Exchange servers found matching criteria")
		except Exception as e:
			logging.error(f"Error in get_exchange_server: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def get_netshare(
		computer: str = "",
		computername: str = ""
	) -> str:
		"""
		Enumerates shares on a specified computer.

		Args:
			computer: Target computer hostname or IP address.
			computername: Alias for computer. Specify only one of computer or computername.
		"""
		target_computer = computer or computername
		if not target_computer:
			return _format_mcp_response(success=False, message="Either 'computer' or 'computername' must be specified")
		if computer and computername:
			 return _format_mcp_response(success=False, message="Specify only one of 'computer' or 'computername'")

		try:
			args = type('Args', (), {
				'computer': computer if computer else None,
				'computername': computername if computername else None,
				'tableview': '',
				'count': False,
				'server': None,
				'outfile': None,
				'module': 'Get-NetShare'
			})
			result = powerview_instance.get_netshare(args=args)
			return _format_mcp_response(data=result, message=f"No shares found on {target_computer}")
		except Exception as e:
			error_msg = str(e)
			if "rpc_s_access_denied" in error_msg.lower():
				error_msg = f"Access denied to {target_computer}. Check permissions."
			elif "connection" in error_msg.lower() or "Errno 111" in error_msg:
				error_msg = f"Could not connect to {target_computer}. Check network connectivity and RPC/SMB services."
			logging.error(f"Error in get_netshare: {str(e)}")
			return _format_mcp_response(error=error_msg)

	@mcp.tool()
	async def set_domain_user_password(
		identity: str,
		accountpassword: str,
		oldpassword: str | None = None
	) -> str:
		"""
		Sets the password for a specified domain user.

		Args:
			identity: The identity (e.g., sAMAccountName) of the user account.
			accountpassword: The new password to set for the user account.
			oldpassword: The current password (required for non-admin password changes).
		"""
		if not identity or not accountpassword:
				return _format_mcp_response(success=False, message="Identity and AccountPassword are required.")

		try:
			args = type('Args', (), {
				'identity': identity,
				'accountpassword': accountpassword,
				'oldpassword': oldpassword,
				'server': None,
				'outfile': None,
				'module': 'Set-DomainUserPassword'
			})

			result = powerview_instance.set_domainuserpassword(
				identity=identity,
				accountpassword=accountpassword,
				oldpassword=oldpassword,
				args=args
			)

			if result is True:
				return _format_mcp_response(success=True, message=f"Password for '{identity}' set successfully.")
			elif result is False:
				error_detail = f"Failed to set password for '{identity}'. Check permissions, password complexity/history, or if the old password is required/correct."
				logging.warning(f"set_domain_user_password returned False for '{identity}'.")
				return _format_mcp_response(success=False, message=error_detail)
			else: # Assumes None means user not found/ambiguous
				error_detail = f"User identity '{identity}' not found or multiple users matched."
				logging.error(f"set_domain_user_password could not find unique user for '{identity}'.")
				return _format_mcp_response(success=False, message=error_detail)

		except Exception as e:
			logging.error(f"Exception in set_domain_user_password for '{identity}': {str(e)}")
			error_msg = str(e)
			return _format_mcp_response(error=f"An unexpected exception occurred: {error_msg}")

	@mcp.tool()
	async def add_domain_user(
		username: str,
		userpass: str,
		basedn: str | None = None
	) -> str:
		"""
		Adds a new domain user.

		Args:
			username: The sAMAccountName for the new user.
			userpass: The password for the new user.
			basedn: The distinguished name of the container/OU to add the user to (default: CN=Users,<domainDN>).
		"""
		if not username or not userpass:
			return _format_mcp_response(success=False, message="Username and Userpass are required.")
		try:
			args = type('Args', (), {'module': 'Add-DomainUser'})
			result = powerview_instance.add_domainuser(
				username=username,
				userpass=userpass,
				basedn=basedn,
				args=args
			)
			if result: # Assuming result is DN on success
				return _format_mcp_response(success=True, message=f"User '{username}' added successfully.", data={"dn": result})
			else:
				logging.warning(f"add_domain_user returned unexpected value for '{username}'.")
				return _format_mcp_response(success=False, message=f"Failed to add user '{username}'. Check permissions or if user already exists.")
		except Exception as e:
			logging.error(f"Exception in add_domain_user for '{username}': {str(e)}")
			return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def remove_domain_user(
		identity: str
	) -> str:
		"""
		Removes a domain user.

		Args:
			identity: The identity (e.g., sAMAccountName, DN) of the user to remove.
		"""
		if not identity:
			return _format_mcp_response(success=False, message="Identity is required.")
		try:
			result = powerview_instance.remove_domainuser(identity=identity)
			if result is True:
				return _format_mcp_response(success=True, message=f"User '{identity}' removed successfully.")
			else:
				msg = f"Failed to remove user '{identity}'. User might not exist or check permissions."
				logging.warning(f"remove_domain_user returned non-True for '{identity}'.")
				return _format_mcp_response(success=False, message=msg)
		except Exception as e:
			logging.error(f"Exception in remove_domain_user for '{identity}': {str(e)}")
			return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def add_domain_group_member(
		identity: str,
		members: str
	) -> str:
		"""
		Adds one or more members to a domain group.

		Args:
			identity: The identity (e.g., sAMAccountName, DN) of the group.
			members: One or more member identities (sAMAccountName, DN) to add, separated by commas. Ensure full DNs are used if needed.
		"""
		if not identity or not members:
			return _format_mcp_response(success=False, message="Identity (group) and Members are required.")
		try:
			if not members.strip():
				return _format_mcp_response(success=False, message="Members string cannot be empty.")
			
			# Split comma-separated members into a list
			member_list = [m.strip() for m in members.split(',') if m.strip()]

			for member in member_list:
				if not member.strip():
					return _format_mcp_response(success=False, message=f"Member '{member}' is empty.")

				result = powerview_instance.add_domaingroupmember(identity=identity, members=member, args=None)

			if result is True:
					return _format_mcp_response(success=True, message=f"Attempted to add members {member_list} to group '{identity}'.")
			else:
					msg = f"Operation completed for group '{identity}' with members {member_list}. Some members might not have been added (check permissions, existence)."
					logging.warning(f"add_domaingroupmember returned non-True for group '{identity}'.")
					# Use status_override to indicate potential partial failure
					return _format_mcp_response(success=False, message=msg, status_override="partial_failure")
		except Exception as e:
			logging.error(f"Exception in add_domaingroupmember for group '{identity}': {str(e)}")
			return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def remove_domain_group_member(
		identity: str,
		members: str
	) -> str:
		"""
		Removes one or more members from a domain group.

		Args:
			identity: The identity (e.g., sAMAccountName, DN) of the group.
			members: One or more member identities (sAMAccountName, DN) to remove, separated by commas. Ensure full DNs are used if needed.
		"""
		if not identity or not members:
			return _format_mcp_response(success=False, message="Identity (group) and Members are required.")
		try:
			if not members.strip():
				return _format_mcp_response(success=False, message="Members string cannot be empty.")
			# Split comma-separated members into a list
			member_list = [m.strip() for m in members.split(',') if m.strip()]

			result = powerview_instance.remove_domaingroupmember(identity=identity, members=member_list, args=None)

			if result is True:
					return _format_mcp_response(success=True, message=f"Attempted to remove members {member_list} from group '{identity}'.")
			else:
					msg = f"Operation completed for group '{identity}' with members {member_list}. Some members might not have been removed (check permissions, existence)."
					logging.warning(f"remove_domaingroupmember returned non-True for group '{identity}'.")
					return _format_mcp_response(success=False, message=msg, status_override="partial_failure")
		except Exception as e:
			logging.error(f"Exception in remove_domaingroupmember for group '{identity}': {str(e)}")
			return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def add_domain_object_acl(
		targetidentity: str,
		principalidentity: str,
		rights: str = "fullcontrol",
		rights_guid: str | None = None,
		ace_type: str = "allowed",
		inheritance: bool = False
	) -> str:
		"""
		Adds an Access Control Entry (ACE) to a domain object's ACL.

		Args:
			targetidentity: Identity of the target object (user, group, computer, etc.). Can be a name (sAMAccountName, UPN) or distinguishedName (DN). If a name resolves to multiple objects, specify the DN instead.
			principalidentity: Identity of the principal (user, group) being granted rights. Can be a name (sAMAccountName, UPN) or distinguishedName (DN). If a name resolves to multiple objects, specify the DN instead.
			rights: The rights to grant. Defaults to 'fullcontrol'. Common examples include 'fullcontrol', 'genericall', 'genericwrite', 'writeowner', 'writedacl', 'writeproperty', 'self', 'rp', 'wp', 'cr', 'lc', 'cc', 'rc', 'lo', 'dt', 'wd', 'wo'. Parser-specific choices: ['immutable', 'resetpassword', 'writemembers', 'dcsync'].
			rights_guid: GUID string for specific extended rights or property sets.
			ace_type: Type of ACE ('allowed' or 'denied'). Defaults to 'allowed'.
			inheritance: Apply inheritance flags (True/False). Defaults to False.
		"""
		if not targetidentity or not principalidentity:
				return _format_mcp_response(success=False, message="TargetIdentity and PrincipalIdentity are required.")
		try:
			result = powerview_instance.add_domainobjectacl(
				targetidentity=targetidentity,
				principalidentity=principalidentity,
				rights=rights,
				rights_guid=rights_guid,
				ace_type=ace_type,
				inheritance=inheritance
			)
			if result is True:
				return _format_mcp_response(success=True, message=f"Successfully added '{rights}' ACE for '{principalidentity}' on '{targetidentity}'.")
			else:
				msg = f"Failed to add ACE for '{principalidentity}' on '{targetidentity}'. Check inputs and permissions."
				logging.warning(f"add_domain_object_acl returned non-True for target '{targetidentity}'.")
				return _format_mcp_response(success=False, message=msg)
		except Exception as e:
			logging.error(f"Exception in add_domain_object_acl for target '{targetidentity}': {str(e)}")
			return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def remove_domain_object_acl(
		targetidentity: str,
		principalidentity: str,
		rights: str = "fullcontrol",
		rights_guid: str | None = None,
		ace_type: str = "allowed",
		inheritance: bool = False
	) -> str:
		"""
		Removes an Access Control Entry (ACE) from a domain object's ACL.

		Args:
			targetidentity: Identity of the target object (user, group, computer, etc.).
			principalidentity: Identity of the principal (user, group) whose rights are being removed.
			rights: The rights to remove. Must match the ACE being removed. Defaults to 'fullcontrol'. Common examples are the same as for adding ACEs. Parser-specific choices: ['immutable', 'resetpassword', 'writemembers', 'dcsync'].
			rights_guid: GUID string if removing specific extended rights or property sets.
			ace_type: Type of ACE ('allowed' or 'denied'). Defaults to 'allowed'.
			inheritance: Match inheritance flags (True/False). Defaults to False.
		"""
		if not targetidentity or not principalidentity:
				return _format_mcp_response(success=False, message="TargetIdentity and PrincipalIdentity are required.")
		try:
			result = powerview_instance.remove_domainobjectacl(
				targetidentity=targetidentity,
				principalidentity=principalidentity,
				rights=rights,
				rights_guid=rights_guid,
				ace_type=ace_type,
				inheritance=inheritance
			)
			if result is True:
				return _format_mcp_response(success=True, message=f"Successfully removed '{rights}' ACE for '{principalidentity}' from '{targetidentity}'.")
			else:
				msg = f"Failed to remove ACE for '{principalidentity}' from '{targetidentity}'. Check if ACE exists with specified parameters and check permissions."
				logging.warning(f"remove_domain_object_acl returned non-True for target '{targetidentity}'.")
				return _format_mcp_response(success=False, message=msg)
		except Exception as e:
			logging.error(f"Exception in remove_domain_object_acl for target '{targetidentity}': {str(e)}")
			return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def set_domain_object_owner(
		targetidentity: str,
		principalidentity: str,
		searchbase: str | None = None
	) -> str:
		"""
		Sets the owner for a specified domain object.

		Args:
			targetidentity: Identity of the target object (user, group, computer, etc.).
			principalidentity: Identity of the principal (user, group) to set as the new owner.
			searchbase: Specify the search base DN for the target object.
		"""
		if not targetidentity or not principalidentity:
				return _format_mcp_response(success=False, message="TargetIdentity and PrincipalIdentity are required.")
		try:
			result = powerview_instance.set_domainobjectowner(
				targetidentity=targetidentity,
				principalidentity=principalidentity,
				searchbase=searchbase,
				args=None # Args not used by underlying function for core logic
			)
			if result is True:
					return _format_mcp_response(success=True, message=f"Successfully set '{principalidentity}' as owner of '{targetidentity}'.")
			else:
					msg = f"Failed to set owner for '{targetidentity}'. Check identities and permissions."
					logging.warning(f"set_domain_object_owner returned non-True for target '{targetidentity}'.")
					return _format_mcp_response(success=False, message=msg)
		except Exception as e:
				logging.error(f"Exception in set_domain_object_owner for target '{targetidentity}': {str(e)}")
				return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def set_domain_computer_password(
		identity: str,
		accountpassword: str,
		oldpassword: str | None = None
	) -> str:
		"""
		Sets the password for a specified domain computer account.

		Args:
			identity: The identity (e.g., sAMAccountName) of the computer account.
			accountpassword: The new password to set for the computer account.
			oldpassword: The current password (might be required in some contexts).
		"""
		if not identity or not accountpassword:
			return _format_mcp_response(success=False, message="Identity and AccountPassword are required.")
		try:
				args = type('Args', (), {
				'identity': identity,
				'accountpassword': accountpassword,
				'oldpassword': oldpassword,
				'server': None,
				'outfile': None,
				'module': 'Set-DomainComputerPassword'
				})
				result = powerview_instance.set_domaincomputerpassword(
					identity=identity,
					accountpassword=accountpassword,
					oldpassword=oldpassword,
					args=args
				)
				if result is True:
					return _format_mcp_response(success=True, message=f"Password for computer '{identity}' set successfully.")
				else:
					error_detail = f"Failed to set password for computer '{identity}'. Check permissions or identity."
					logging.warning(f"set_domain_computer_password returned non-True for '{identity}'.")
					return _format_mcp_response(success=False, message=error_detail)
		except Exception as e:
				logging.error(f"Exception in set_domain_computer_password for '{identity}': {str(e)}")
				return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def convert_sid_to_name(
		sid: str
	) -> str:
		"""Convert a SID to a name."""
		try:
			result = powerview_instance.convertfrom_sid(sid)
			return _format_mcp_response(data={"name": result})
		except Exception as e:
			logging.error(f"Exception in convert_sid_to_name for '{sid}': {str(e)}")
			return _format_mcp_response(error=f"An unexpected exception occurred: {str(e)}")

	@mcp.tool()
	async def get_current_auth_context() -> str:
		"""Get the current authenticated user context for the PowerView session."""
		try:
			identity = powerview_instance.conn.who_am_i()
			username = identity.split("\\")[-1] if "\\" in identity else identity
			return _format_mcp_response(data={"identity": username})
		except AttributeError:
			logging.error("Error in get_current_auth_context: powerview_instance or connection object not available.")
			return _format_mcp_response(error="Internal server error: Could not access connection details.")
		except Exception as e:
			logging.error(f"Error in get_current_auth_context: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def generate_findings_report(
		custom_findings: str,
		report_title: str = "PowerView Findings Report",
		target_domain: str = "",
		report_format: str = "markdown"
	) -> str:
		"""Generate a custom security findings report based on user-defined findings.
		
		Args:
			custom_findings: JSON string containing custom finding dictionaries with keys like "Title", "Description", "Risk", "Recommendation" etc.
			report_title: Title for the report
			target_domain: Target domain name for the report context
			report_format: Output format (markdown, json, or text)
		"""
		try:
			# Parse the custom findings from JSON string
			findings_list = json.loads(custom_findings)
			if not isinstance(findings_list, list):
				findings_list = [findings_list]  # Convert single dict to list
				
			# Initialize report data
			timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
			
			# Process each custom finding dict and normalize keys
			normalized_findings = []
			for finding in findings_list:
				# Normalize keys (case-insensitive matching)
				normalized = {}
				key_map = {
					"title": ["title", "name", "finding", "issue"],
					"risk": ["risk", "severity", "level", "criticality"],
					"description": ["description", "desc", "details", "info"],
					"affected_entity": ["affected_entity", "affected_objects", "affected", "entity", "entities", "target"],
					"recommendation": ["recommendation", "remediation", "fix", "solution", "mitigation"],
					"notes": ["notes", "comments", "additional", "context"]
				}
				
				# Try to map keys from the input dict to our normalized structure
				for norm_key, possible_keys in key_map.items():
					for pk in possible_keys:
						for k in finding.keys():
							if k.lower() == pk.lower():
								normalized[norm_key] = finding[k]
								break
						if norm_key in normalized:
							break
						
				# Ensure required fields exist with defaults if missing
				if "title" not in normalized:
					normalized["title"] = "Unnamed Finding"
				if "risk" not in normalized:
					normalized["risk"] = "Medium"  # Default risk level
				if "description" not in normalized:
					normalized["description"] = "No description provided."
				if "recommendation" not in normalized:
					normalized["recommendation"] = "No specific recommendation provided."

				# Handle affected_entity field specially (could be string or list or dict)
				if "affected_entity" in normalized:
					if isinstance(normalized["affected_entity"], str):
						normalized["affected_objects"] = [{"name": normalized["affected_entity"]}]
					elif isinstance(normalized["affected_entity"], list):
						if all(isinstance(item, str) for item in normalized["affected_entity"]):
							normalized["affected_objects"] = [{"name": item} for item in normalized["affected_entity"]]
						else:
							normalized["affected_objects"] = normalized["affected_entity"]
					elif isinstance(normalized["affected_entity"], dict):
						normalized["affected_objects"] = [normalized["affected_entity"]]
					normalized["count"] = len(normalized["affected_objects"])
				else:
					normalized["affected_objects"] = []
					normalized["count"] = 0
					
				# Add any extra fields not in our mapping
				for k, v in finding.items():
					if not any(k.lower() == pk.lower() for norm_keys in key_map.values() for pk in norm_keys):
						normalized[k] = v
						
				normalized_findings.append(normalized)
				
			# Format the report
			report = {
				"title": report_title,
				"target_domain": target_domain,
				"timestamp": timestamp,
				"summary": {
					"total_findings": len(normalized_findings),
					"risk_breakdown": {
						"Critical": len([f for f in normalized_findings if f["risk"].lower() == "critical"]),
						"High": len([f for f in normalized_findings if f["risk"].lower() == "high"]),
						"Medium": len([f for f in normalized_findings if f["risk"].lower() == "medium"]),
						"Low": len([f for f in normalized_findings if f["risk"].lower() == "low"])
					}
				},
				"findings": normalized_findings
			}
			
			# Generate report based on requested format
			if report_format.lower() == "json":
				return _format_mcp_response(data=report)
			elif report_format.lower() == "markdown":
				md_report = f"# {report_title}\n\n"
				md_report += f"**Target Domain:** {target_domain}\n"
				md_report += f"**Report Generated:** {timestamp}\n\n"
				
				md_report += "## Summary\n\n"
				md_report += f"**Total Findings:** {report['summary']['total_findings']}\n\n"
				md_report += "**Risk Breakdown:**\n"
				md_report += f"- Critical: {report['summary']['risk_breakdown']['Critical']}\n"
				md_report += f"- High: {report['summary']['risk_breakdown']['High']}\n"
				md_report += f"- Medium: {report['summary']['risk_breakdown']['Medium']}\n"
				md_report += f"- Low: {report['summary']['risk_breakdown']['Low']}\n\n"
				
				if normalized_findings:
					md_report += "## Findings\n\n"
					for i, finding in enumerate(normalized_findings, 1):
						md_report += f"### {i}. {finding['title']} ({finding['risk']})\n\n"
						md_report += f"**Description:** {finding['description']}\n\n"
						
						if finding.get('affected_objects'):
							md_report += f"**Affected Objects:** {finding['count']}\n\n"
							md_report += "**Examples:**\n\n"
							md_report += "```\n"
							for obj in finding['affected_objects'][:5]:  # Limit to 5 examples
								md_report += json.dumps(obj, indent=2) + "\n"
							md_report += "```\n\n"
						
						md_report += f"**Recommendation:** {finding['recommendation']}\n\n"
						
						# Add notes if present
						if 'notes' in finding:
							md_report += f"**Notes:** {finding['notes']}\n\n"
							
						# Add any custom fields
						custom_fields = [k for k in finding.keys() if k not in ["title", "risk", "description", "affected_objects", "count", "recommendation", "notes"]]
						if custom_fields:
							md_report += "**Additional Information:**\n\n"
							for field in custom_fields:
								md_report += f"- **{field}:** {finding[field]}\n"
							md_report += "\n"
						
						md_report += "---\n\n"
				else:
					md_report += "## No Findings\n\n"
					md_report += "No security findings were provided.\n\n"
				
				return _format_mcp_response(data={"report": md_report})
			else:  # text format
				text_report = f"{report_title}\n"
				text_report += f"Target Domain: {target_domain}\n"
				text_report += f"Report Generated: {timestamp}\n\n"
				
				text_report += "SUMMARY:\n"
				text_report += f"Total Findings: {report['summary']['total_findings']}\n\n"
				text_report += "Risk Breakdown:\n"
				text_report += f"Critical: {report['summary']['risk_breakdown']['Critical']}\n"
				text_report += f"High: {report['summary']['risk_breakdown']['High']}\n"
				text_report += f"Medium: {report['summary']['risk_breakdown']['Medium']}\n"
				text_report += f"Low: {report['summary']['risk_breakdown']['Low']}\n\n"
				
				if normalized_findings:
					text_report += "FINDINGS:\n\n"
					for i, finding in enumerate(normalized_findings, 1):
						text_report += f"{i}. {finding['title']} ({finding['risk']})\n\n"
						text_report += f"Description: {finding['description']}\n\n"
						
						if finding.get('affected_objects'):
							text_report += f"Affected Objects: {finding['count']}\n\n"
						
						text_report += f"Recommendation: {finding['recommendation']}\n\n"
						
						# Add notes if present
						if 'notes' in finding:
							text_report += f"Notes: {finding['notes']}\n\n"
						
						text_report += "----------\n\n"
				else:
					text_report += "NO FINDINGS\n\n"
					text_report += "No security findings were provided.\n\n"
				
				return _format_mcp_response(data={"report": text_report})
				
		except Exception as e:
			logging.error(f"Error in generate_custom_findings_report: {str(e)}")
			import traceback
			trace = traceback.format_exc()
			return _format_mcp_response(error=f"Failed to generate report: {str(e)}\n{trace}")

	@mcp.tool()
	async def find_localadminaccess(
		computer: str = None,
		username: str = None,
		password: str = None,
		nthash: str = None,
		lmhash: str = None,
		no_cache: bool = False,
	) -> str:
		"""
		Enumerate computers where the current user has local admin access. Accepts a single computer or checks all domain computers if not specified. When computer is not specified, it will check all domain computers (careful because you might be dealing with thousands of hosts).
		
		Args:
			computer: The computer to check for local admin access. Leave blank to check all domain computers.
			username: The username to use for authentication. Ignore to use current user context.
			password: The password to use for authentication. Ignore to use current user context.
			nthash: The NTHash to use for authentication. Ignore to use current user context.
			lmhash: The LMHash to use for authentication. Ignore to use current user context.
			no_cache: Whether to use cached results.
		"""
		try:
			args = type('Args', (), {
				'computer': computer,
				'username': username,
				'password': password,
				'nthash': nthash,
				'lmhash': lmhash,
				'no_cache': no_cache,
				'module': 'Find-LocalAdminAccess'
			})
			result = powerview_instance.find_localadminaccess(args=args)
			return _format_mcp_response(data=result, message="No local admin access found on any host.")
		except Exception as e:
			logging.error(f"Error in find_localadminaccess: {str(e)}")
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def smb_connect(
		computer: str,
		username: str = "",
		password: str = "",
		nthash: str = "",
		lmhash: str = "",
		domain: str = ""
	) -> str:
		"""
		Establish an SMB connection to target system.
		
		This function establishes a connection to a remote system using SMB protocol, 
		which can be leveraged for file operations, privilege escalation, lateral movement,
		or intelligence gathering. Supports multiple authentication methods including
		NTLM hash pass-the-hash attacks.
		
		Parameters:
			computer: Target hostname or IP address
			username: Optional: Account username (optional if using current context)
			password: Optional: Account password in cleartext
			nthash: Optional: NT hash for pass-the-hash attacks
			lmhash: Optional: LM hash (default: None)
			domain: Optional: Domain name (will use current domain if not specified)
		
		Returns:
			Connection status information
		"""
		try:
			if username and ('/' in username or '\\' in username):
				domain, username = username.replace('/', '\\').split('\\')
			
			if not computer:
				return _format_mcp_response(error="Computer name/IP is required")

			is_fqdn = False
			host = ""

			if not is_ipaddress(computer):
				is_fqdn = True
				if not is_valid_fqdn(computer):
					host = f"{computer}.{powerview_instance.domain}"
				else:
					host = computer
			else:
				host = computer

			if powerview_instance.use_kerberos:
				if is_ipaddress(computer):
					return _format_mcp_response(error="FQDN must be used for kerberos authentication")
			else:
				if is_fqdn:
					host = host2ip(host, powerview_instance.nameserver, 3, True, 
							use_system_ns=powerview_instance.use_system_nameserver)

			if not host:
				return _format_mcp_response(error="Host not found")

			client = powerview_instance.conn.init_smb_session(
				host,
				username=username,
				password=password,
				nthash=nthash,
				lmhash=lmhash,
				domain=domain
			)
			
			if not client:
				return _format_mcp_response(error=f"Failed to connect to {host}")

			return _format_mcp_response(data={"status": "connected", "host": host})
		except Exception as e:
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def smb_shares(
		computer: str
	) -> str:
		"""List available SMB shares on a target system.
		
		Enumerates available SMB shares on a target system, useful for reconnaissance
		and identifying valuable data.
		
		Parameters:
			computer: Target hostname or IP address
		
		Returns:
			List of available SMB shares of the target computer
		"""
		try:
			host = computer.lower()
			
			if not host:
				return _format_mcp_response(error="Computer name/IP is required")
			
			try:
				client = powerview_instance.conn.init_smb_session(host)
			except Exception as e:
				return _format_mcp_response(error="No active SMB session. Please connect first")

			smb_client = SMBClient(client)
			shares = smb_client.shares()

			formatted_shares = []
			for share in shares:
				entry = {
					"Name": share['shi1_netname'][:-1],
					"Remark": share['shi1_remark'][:-1],
					"Address": host
				}
				formatted_shares.append({"attributes": entry})

			return _format_mcp_response(data=formatted_shares)
		except Exception as e:
			return _format_mcp_response(error=str(e))

	@mcp.tool()
	async def smb_ls(
		computer: str,
		share: str,
		path: str = ""
	) -> str:
		"""List contents of a directory on a remote SMB share.
		
		Enumerates files and directories within a specified share path, useful for
		reconnaissance, identifying valuable data, and planning exfiltration.
		Directory traversal is supported to explore the target filesystem.
		
		Parameters:
			computer: Target hostname or IP address
			share: Share name to access (e.g., "C$", "ADMIN$", "NETLOGON")
			path: Directory path within the share to list (default: root of share) (i.e.: "\\Users\\Public\\Desktop")
		
		Returns:
			File listing with metadata (size, timestamps, attributes)
		"""
		try:
			host = computer.lower()
			
			if not host or not share:
				return _format_mcp_response(error="Computer name/IP and share name are required")

			try:
				client = powerview_instance.conn.init_smb_session(host)
			except Exception as e:
				return _format_mcp_response(error="No active SMB session. Please connect first")
			
			smb_client = SMBClient(client)
			
			files = smb_client.ls(share, path)
			logging.debug(f"[SMB LS] Listing {path} on {host} with share {share}")
			
			file_list = []
			for f in files:
				name = f.get_longname()
				if name in ['.', '..']:
					continue
				
				file_info = {
					"name": name,
					"size": f.get_filesize(),
					"is_directory": f.is_directory(),
					"created": str(f.get_ctime()),
					"modified": str(f.get_mtime()),
					"accessed": str(f.get_atime())
				}
				file_list.append(file_info)

			return _format_mcp_response(data=file_list)
		except Exception as e:
			logging.error(f"[SMB LS] Error: {str(e)}")
			return _format_mcp_response(error=str(e))


from .. import ENUMERATION
from ldap3 import DSA, SCHEMA, ALL, BASE, SUBTREE, SEQUENCE_TYPES
from ldap3.protocol.rfc4512 import SchemaInfo, DsaInfo
from ldap3.protocol.formatters.standard import format_attribute_values

class Server(object):
    def __init__(self, host, ssl=False, port=9389, get_info=ALL, formatter=None, validator=None, resource=ENUMERATION):
        self.host = host
        self.port = port
        self.ssl = ssl
        self.custom_formatter = formatter
        self.custom_validator = validator
        self.resource = resource
        self.get_info = get_info
        self._dsa_info = None
        self._schema_info = None
    
    def get_info_from_server(self, connection):
        """
        reads info from DSE and from subschema
        """
        if connection:
            if self.get_info in [DSA, ALL]:
                self._get_dsa_info(connection)
            if self.get_info in [SCHEMA, ALL]:
                self._get_schema_info(connection)

    def _get_dsa_info(self, connection):
        """
        Retrieve DSE operational attribute as per RFC4512 (5.1) via ADWS.
        Querying RootDSE (empty base, base scope) for server info.
        """
        # Attributes commonly available via ADWS RootDSE
        # Requesting only attributes essential for PowerView initialization
        attributes_to_request = [
            'rootDomainNamingContext',       # Needed for forest_dn
            'defaultNamingContext',        # Needed for root_dn
            'configurationNamingContext',  # Needed for configuration_dn
            'ldapServiceName',           # Needed for flatName
            'dnsHostName',               # Needed for dc_dnshostname
            'supportedControl',          # Removed - less likely to be essential for init
            'supportedExtension',
            'supportedCapabilities',
            'supportedLDAPVersion',
            'supportedSASLMechanisms',
            'subschemaSubentry',         # Needed for _get_schema_info, but not init
            'serverName',
            'domainControllerFunctionality',
            'domainFunctionality',
            'forestFunctionality',
            'highestCommittedUSN',
            'isSynchronized',
            'isGlobalCatalogReady',
            'currentTime'
        ]
        response = connection.search(search_base='',
                                   search_filter='(objectClass=*)',
                                   search_scope=BASE,
                                   attributes=attributes_to_request)
        self._dsa_info = DsaInfo(response[0]['attributes'], response[0]['raw_attributes'])

    def _get_schema_info(self, connection, entry=''):
        """
        Retrieve schema from subschemaSubentry DSE attribute, per RFC
        4512 (4.4 and 5.1); entry = '' means DSE via ADWS.
        """
        schema_entry = None
        if self._dsa_info and entry == '':  # subschemaSubentry already present in dsaInfo
            if isinstance(self._dsa_info.schema_entry, SEQUENCE_TYPES):
                schema_entry = self._dsa_info.schema_entry[0] if self._dsa_info.schema_entry else None
            else:
                schema_entry = self._dsa_info.schema_entry if self._dsa_info.schema_entry else None
        else:
            result = connection.search(entry, '(objectClass=*)', BASE, attributes=['subschemaSubentry'], get_operational_attributes=True)
            if result and 'subschemaSubentry' in result[0]['raw_attributes']:
                if len(result[0]['raw_attributes']['subschemaSubentry']) > 0:
                    schema_entry = result[0]['raw_attributes']['subschemaSubentry'][0]
        
        if schema_entry:
            response = connection.search(schema_entry,
                                        search_filter='(objectClass=subschema)',
                                        search_scope=BASE,
                                        attributes=[
                                            'objectClasses',
                                            'attributeTypes',
                                            'createTimestamp',
                                            'modifyTimestamp'
                                        ])
            self._schema_info = SchemaInfo(schema_entry, response[0]['attributes'], response[0]['raw_attributes'])
            if self._schema_info:  # if schema is valid tries to apply formatter to the "other" dict with raw values for schema and info
                for attribute in self._schema_info.other:
                    self._schema_info.other[attribute] = format_attribute_values(self._schema_info, attribute, self._schema_info.raw[attribute], self.custom_formatter)
                if self._dsa_info:  # try to apply formatter to the "other" dict with dsa info raw values
                    for attribute in self._dsa_info.other:
                        self._dsa_info.other[attribute] = format_attribute_values(self._schema_info, attribute, self._dsa_info.raw[attribute], self.custom_formatter)

    @property
    def info(self):
        return self._dsa_info

    @property
    def schema(self):
        return self._schema_info


#!/usr/bin/env python3
import random

from ..core.constants import (
	WILDCARD, EXACT_MATCH_ATTRIBUTES, WILDCARD_SAFE_ATTRIBUTES
)
from ..utils.helpers import in_exception, LdapObfuscate
from ..utils.validators import is_sid_attribute

def wildcard_expansion_obfuscation(parsed_structure, transform_func):
	"""
	Expand wildcards with additional characters for evasion.
	Following pattern of safe wildcard transformations.
	"""
	def obfuscate_wildcards(attr, operator, value):
		if in_exception(attr) or value == WILDCARD:
			return attr, value
		
		if LdapObfuscate.is_number(value):
			return attr, value
		
		if attr.lower() in EXACT_MATCH_ATTRIBUTES:
			return attr, value
			
		if is_sid_attribute(attr):
			return attr, value
		
		if operator == "=" and value and WILDCARD not in value:
			if attr.lower() in WILDCARD_SAFE_ATTRIBUTES:
				if random.choice([True, False]):
					if random.choice([True, False]):
						value = WILDCARD + value
					if random.choice([True, False]):
						value = value + WILDCARD
		
		return attr, value

	transform_func(parsed_structure, obfuscate_wildcards) 


NAMESPACES = {
    "ad": "http://schemas.microsoft.com/2008/1/ActiveDirectory",
    "addata": "http://schemas.microsoft.com/2008/1/ActiveDirectory/Data",
    "adlq": "http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/LdapQuery",
    "da": "http://schemas.microsoft.com/2006/11/IdentityManagement/DirectoryAccess",
    "ca": "http://schemas.microsoft.com/2008/1/ActiveDirectory/CustomActions",

    "soapenv": "http://www.w3.org/2003/05/soap-envelope",
    "wsa": "http://www.w3.org/2005/08/addressing",
    "wsen": "http://schemas.xmlsoap.org/ws/2004/09/enumeration",
    "wxf": "http://schemas.xmlsoap.org/ws/2004/09/transfer",
    "xsd": "http://www.w3.org/2001/XMLSchema",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",

    "s": "http://www.w3.org/2003/05/soap-envelope",
    "a": "http://www.w3.org/2005/08/addressing",
}

LDAP_QUERY_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://www.w3.org/2005/08/addressing"
    xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
    xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <s:Header>
        <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate</a:Action>
        <ad:instance>ldap:389</ad:instance>
        <a:MessageID>urn:uuid:{uuid}</a:MessageID>
        <a:ReplyTo>
            <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
        </a:ReplyTo>
        <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Enumeration</a:To>
    </s:Header>
    <s:Body xmlns:wsen="http://schemas.xmlsoap.org/ws/2004/09/enumeration"
        xmlns:adlq="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/LdapQuery">
        <wsen:Enumerate>
            <wsen:Filter Dialect="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/LdapQuery">
                <adlq:LdapQuery>
                    <adlq:Filter>{query}</adlq:Filter>
                    <adlq:BaseObject>{search_base}</adlq:BaseObject>
                    <adlq:Scope>{search_scope}</adlq:Scope>
                </adlq:LdapQuery>
            </wsen:Filter>
            <ad:Selection Dialect="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/XPath-Level-1">
                <ad:SelectionProperty>ad:distinguishedName</ad:SelectionProperty>
                {attributes}
            </ad:Selection>
        </wsen:Enumerate>
    </s:Body>
</s:Envelope>"""

LDAP_PULL_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://www.w3.org/2005/08/addressing"
    xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
    xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <s:Header>
        <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/enumeration/Pull</a:Action>
        <ad:instance>ldap:389</ad:instance>
        <a:MessageID>urn:uuid:{uuid}</a:MessageID>
        <a:ReplyTo>
            <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
        </a:ReplyTo>
        <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Enumeration</a:To>
    </s:Header>
    <s:Body xmlns:wsen="http://schemas.xmlsoap.org/ws/2004/09/enumeration">
        <wsen:Pull>
            <wsen:EnumerationContext>{enum_ctx}</wsen:EnumerationContext>
            <wsen:MaxElements>256</wsen:MaxElements>
        </wsen:Pull>
    </s:Body>
</s:Envelope>"""


LDAP_PUT_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
        xmlns:a="http://www.w3.org/2005/08/addressing"
        xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
        xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
        xmlns:xsd="http://www.w3.org/2001/XMLSchema"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        xmlns:da="http://schemas.microsoft.com/2006/11/IdentityManagement/DirectoryAccess">
        <s:Header>
            <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/transfer/Put</a:Action>
            <ad:instance>ldap:389</ad:instance>
            <ad:objectReferenceProperty>{object_ref}</ad:objectReferenceProperty>
            <da:IdentityManagementOperation s:mustUnderstand="1"
                xmlns:i="http://www.w3.org/2001/XMLSchema-instance"></da:IdentityManagementOperation>
            <a:MessageID>urn:uuid:{uuid}</a:MessageID>
            <a:ReplyTo>
                <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
            </a:ReplyTo>
            <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Resource</a:To>
        </s:Header>
        <s:Body>
            {attributes}
        </s:Body>
    </s:Envelope>"""

LDAP_DELETE_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
            xmlns:a="http://www.w3.org/2005/08/addressing"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xmlns:xsd="http://www.w3.org/2001/XMLSchema"
            xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory">
  <s:Header>
    <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete</a:Action>
    <ad:instance>ldap:389</ad:instance>
    <ad:objectReferenceProperty>{object_ref}</ad:objectReferenceProperty>
    <a:MessageID>urn:uuid:{uuid}</a:MessageID>
    <a:ReplyTo>
      <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
    </a:ReplyTo>
    <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Resource</a:To>
  </s:Header>
  <s:Body>
  </s:Body>
</s:Envelope>
"""

LDAP_ADD_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:wsa="http://www.w3.org/2005/08/addressing">
    <s:Header>
        <wsa:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/transfer/Create</wsa:Action>
        <IdentityManagementOperation s:mustUnderstand="1"
            xmlns="http://schemas.microsoft.com/2006/11/IdentityManagement/DirectoryAccess" />
        <instance xmlns="http://schemas.microsoft.com/2008/1/ActiveDirectory">ldap:389</instance>
        <wsa:MessageID>urn:uuid:{uuid}</wsa:MessageID>
        <wsa:ReplyTo>
            <wsa:Address>http://www.w3.org/2005/08/addressing/anonymous</wsa:Address>
        </wsa:ReplyTo>
        <wsa:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/ResourceFactory</wsa:To>
    </s:Header>
    <s:Body>
        <AddRequest Dialect="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/XPath-Level-1"
            xmlns="http://schemas.microsoft.com/2006/11/IdentityManagement/DirectoryAccess"
            xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
            xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xmlns:xsd="http://www.w3.org/2001/XMLSchema">
            {attributes}
        </AddRequest>
    </s:Body>
</s:Envelope>"""

LDAP_MODIFY_DN_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://www.w3.org/2005/08/addressing"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory">
    <s:Header>
        <a:Action s:mustUnderstand="1">http://schemas.microsoft.com/2008/1/ActiveDirectory/Data/ModifyDNRequest</a:Action>
        <ad:instance>ldap:389</ad:instance>
        <ad:objectReferenceProperty>{object_ref}</ad:objectReferenceProperty>
        <a:MessageID>urn:uuid:{uuid}</a:MessageID>
        <a:ReplyTo>
            <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
        </a:ReplyTo>
        <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Resource</a:To>
    </s:Header>
    <s:Body>
        <da:ModifyDNRequest xmlns:da="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data">
            <da:DistinguishedName>{object_ref}</da:DistinguishedName>
            <da:NewName>{relative_dn}</da:NewName>
            <da:DeleteOldRdn>{delete_old_rdn}</da:DeleteOldRdn>
            <da:NewParentDistinguishedName>{new_superior}</da:NewParentDistinguishedName>
        </da:ModifyDNRequest>
    </s:Body>
</s:Envelope>"""



#!/usr/bin/env python3

import logging

def setup_prompts(mcp):
	"""Register all PowerView prompts with the MCP server."""

	@mcp.prompt()
	async def find_vulnerable_systems(query: str = "") -> str:
		"""Create a prompt to find vulnerable systems."""
		return f"""
		Please analyze the Active Directory environment using PowerView 
		and identify potential security vulnerabilities.
		
		Focus areas:
		- Kerberoastable accounts
		- Accounts with SPNs
		- Privilege escalation paths
		- Resource-Based Constrained Delegation (RBCD) issues
		- Shadow Credentials vulnerabilities
		- Unconstrained delegation
		- Other critical security issues
		
		If there's a specific area to focus on: {query}
		"""

	@mcp.prompt()
	async def ad_mapping_prompt(depth: str = "basic") -> str:
		"""Create a prompt to map the Active Directory environment."""
		detail_level = {
			"basic": "focus on the high-level structure",
			"detailed": "include detailed information about users, groups, and computers",
			"comprehensive": "perform an exhaustive analysis of all AD objects and their relationships"
		}.get(depth.lower(), "focus on the high-level structure")
		
		return f"""
		Please map the Active Directory environment using PowerView tools.
		
		For this mapping, {detail_level}.
		
		Include the following in your analysis:
		- Domain information and properties
		- Domain controllers
		- Organizational Units (OUs) structure
		- Key security groups and their members
		- User accounts of interest
		- Computer accounts and their properties
		- Trust relationships
		- Group Policy Objects (GPOs)
		
		Provide a structured overview of the environment with notable findings.
		"""

	@mcp.prompt()
	async def find_attack_path_from_current_context() -> str:
		"""Create a prompt to find attack paths from the current user context.
		   Note: This prompt cannot dynamically fetch the current user anymore. Execute the tool `get_current_auth_context` to get the current user."""
		current_user_placeholder = "[CURRENT_USER_IDENTITY]"

		return f"""
		Please identify the current authentication context (e.g., using `get_current_auth_context` if available, or based on connection info).
		Let's assume the current user is: {current_user_placeholder}

		Investigate potential attack paths starting from this user context.
		Focus on actions that {current_user_placeholder} might be able to perform, such as:

		1.  Check Group Memberships: Does {current_user_placeholder} belong to any privileged or interesting groups?
			- Use `get_domain_group` with `memberidentity=\'{current_user_placeholder}\'`
		2.  Find Outgoing ACLs: Does {current_user_placeholder} have permissions ON other objects (users, groups, computers)?
			- Use `get_domain_user` or `get_domain_computer` to find the SID for {current_user_placeholder}.
			- Use `get_domain_object_acl` with the SID found.
			- Pay attention to: GenericAll, GenericWrite, WriteOwner, WriteDACL, WriteProperty (especially on group memberships or userAccountControl), User-Force-Change-Password, AddMembers, etc.
		3.  Check for Owned Objects: Does {current_user_placeholder} own any objects?
			- Use `get_domain_object_owner` filtering for the SID of {current_user_placeholder}.
		4.  Check Delegation Rights:
			- Is {current_user_placeholder} configured for Unconstrained Delegation? (`get_domain_user` or `get_domain_computer` with `identity=\'{current_user_placeholder}\'` and check `userAccountControl`)
			- Is {current_user_placeholder} configured for Constrained Delegation (`TRUSTED_TO_AUTH_FOR_DELEGATION`)? (`get_domain_user` or `get_domain_computer` with `identity=\'{current_user_placeholder}\'` and check `msDS-AllowedToDelegateTo`)
			- Is {current_user_placeholder} allowed to delegate TO other services/computers (Resource-Based Constrained Delegation)? (`get_domain_user` or `get_domain_computer` with `identity=\'{current_user_placeholder}\'` and check `msDS-AllowedToActOnBehalfOfOtherIdentity`)
		5.  Check Local Admin Rights: Can {current_user_placeholder} access administrative shares (C$, ADMIN$) or perform administrative actions on any computers? (Requires external tooling or specific RPC calls not directly exposed as simple MCP tools here, e.g., trying `get_netshare` on `\\\\\\\\TARGET\\\\C$`).
		6.  Kerberoasting/ASREPRoasting: Can {current_user_placeholder} request service tickets for accounts with weak passwords or find users vulnerable to ASREPRoasting?
			- Use `invoke_kerberoast`
			- Use `invoke_asreproast`
		7.  Certificate Abuse: Can {current_user_placeholder} request certificates from templates vulnerable to ESC1/ESC4 etc.?
			- Use `get_domain_ca_template` with `vulnerable=True`.

		Prioritize findings based on potential impact (e.g., direct control of admin accounts/groups is higher priority than control of standard users/computers).
		"""



from ...core.connection import Connection
from ldap3 import SUBTREE, DEREF_ALWAYS
from ldap3.utils.dn import safe_dn

def adws_paged_search_generator(connection,
                           search_base,
                           search_filter,
                           search_scope=SUBTREE,
                           dereference_aliases=DEREF_ALWAYS,
                           attributes=None,
                           size_limit=0,
                           time_limit=0,
                           types_only=False,
                           get_operational_attributes=False,
                           controls=None,
                           paged_size=100,
                           paged_criticality=False):
    if connection.check_names and search_base:
        search_base = safe_dn(search_base)

    responses = []
    cookie = None
    total_results = 0
    while True:
        results = connection.search(search_base,
                                   search_filter,
                                   search_scope,
                                   dereference_aliases,
                                   attributes,
                                   size_limit,
                                   time_limit,
                                   types_only,
                                   get_operational_attributes,
                                   controls,
                                   paged_size,
                                   paged_criticality,
                                   None if cookie is True else cookie)
        if not results:
            break
        for entry in results:
            yield entry
            total_results += 1
            if size_limit and total_results >= size_limit:
                return
        if hasattr(connection, 'last_cookie'):
            cookie = connection.last_cookie
        else:
            break
        if not cookie:
            break

def adws_paged_search_accumulator(connection,
                             search_base,
                             search_filter,
                             search_scope=SUBTREE,
                             dereference_aliases=DEREF_ALWAYS,
                             attributes=None,
                             size_limit=0,
                             time_limit=0,
                             types_only=False,
                             get_operational_attributes=False,
                             controls=None,
                             paged_size=100,
                             paged_criticality=False):
    if connection.check_names and search_base:
        search_base = safe_dn(search_base)

    return list(adws_paged_search_generator(connection,
                                           search_base,
                                           search_filter,
                                           search_scope,
                                           dereference_aliases,
                                           attributes,
                                           size_limit,
                                           time_limit,
                                           types_only,
                                           get_operational_attributes,
                                           controls,
                                           paged_size,
                                           paged_criticality))